# Copyright (c) OpenMMLab. All rights reserved.
_base_ = ['../_base_/default_runtime.py']

save_dir = './work_dir/'

model = dict(
    type='DeblurGanV2',
    generator=dict(
        type='DeblurGanV2Generator',
        backbone='FPNMobileNet',
        norm_layer='instance',
        output_ch=3,
        num_filter=64,
        num_filter_fpn=128,
    ),
    discriminator=dict(
        type='DeblurGanV2Discriminator',
        backbone='DoubleGan',
        norm_layer='instance',
        d_layers=3,
    ),
    pixel_loss=dict(
        type='PerceptualLoss', layer_weights={'14': 1}, criterion='mse'),
    disc_loss=dict(type='AdvLoss', loss_type='ragan-ls'),
    adv_lambda=0.001,
    warmup_num=3,
    data_preprocessor=dict(
        type='DataPreprocessor',
        mean=[127.5] * 3,
        std=[127.5] * 3,
    ))

train_pipeline = [
    dict(type='LoadImageFromFile', key='img'),
    dict(type='LoadImageFromFile', key='gt'),
    dict(type='SetValues', dictionary=dict(scale=1)),
    dict(type='PairedAlbuTransForms', size=256, lq_key='img', gt_key='gt'),
    dict(
        type='AlbuCorruptFunction',
        keys=['img'],
        config=[{
            'name': 'cutout',
            'prob': 0.5,
            'num_holes': 3,
            'max_h_size': 25,
            'max_w_size': 25
        }, {
            'name': 'jpeg',
            'quality_lower': 70,
            'quality_upper': 90
        }, {
            'name': 'motion_blur'
        }, {
            'name': 'median_blur'
        }, {
            'name': 'gamma'
        }, {
            'name': 'rgb_shift'
        }, {
            'name': 'hsv_shift'
        }, {
            'name': 'sharpen'
        }]),
    dict(type='PackInputs')
]

val_pipeline = [
    dict(type='LoadImageFromFile', key='img'),
    dict(type='LoadImageFromFile', key='gt'),
    dict(type='SetValues', dictionary=dict(scale=1)),
    dict(type='PairedAlbuTransForms', size=256, lq_key='img', gt_key='gt'),
    dict(
        type='AlbuCorruptFunction',
        keys=['img'],
        config=[{
            'name': 'cutout',
            'prob': 0.5,
            'num_holes': 3,
            'max_h_size': 25,
            'max_w_size': 25
        }, {
            'name': 'jpeg',
            'quality_lower': 70,
            'quality_upper': 90
        }, {
            'name': 'motion_blur'
        }, {
            'name': 'median_blur'
        }, {
            'name': 'gamma'
        }, {
            'name': 'rgb_shift'
        }, {
            'name': 'hsv_shift'
        }, {
            'name': 'sharpen'
        }]),
    dict(type='PackInputs')
]

test_pipeline = [
    dict(type='LoadImageFromFile', key='img'),
    dict(type='LoadImageFromFile', key='gt'),
    dict(type='PackInputs')
]

data_root = 'data/gopro'

train_dataloader = dict(
    batch_size=1,
    num_workers=4,
    persistent_workers=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='gopro', task_name='deblur'),
        data_root=data_root + '/train',
        data_prefix=dict(img='input', gt='target'),
        pipeline=train_pipeline))

val_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='gopro', task_name='deblur'),
        data_root=data_root + '/test',
        data_prefix=dict(img='input', gt='target'),
        pipeline=val_pipeline))

test_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='gopro', task_name='deblur'),
        data_root=data_root + '/test',
        data_prefix=dict(img='input', gt='target'),
        pipeline=test_pipeline))

val_evaluator = dict(
    type='Evaluator', metrics=[
        dict(type='PSNR'),
        dict(type='SSIM'),
    ])

train_cfg = dict(type='EpochBasedTrainLoop', max_epochs=100)
val_cfg = dict(type='MultiValLoop')
test_cfg = dict(type='MultiTestLoop')
test_evaluator = val_evaluator

# optimizer
optim_wrapper = dict(
    constructor='MultiOptimWrapperConstructor',
    generator=dict(
        type='OptimWrapper',
        optimizer=dict(type='Adam', lr=0.0001, betas=(0.5, 0.999))),
    discriminator=dict(
        type='OptimWrapper',
        optimizer=dict(type='Adam', lr=0.0001, betas=(0.5, 0.999))),
)

# learning policy
param_scheduler = dict(
    type='LinearLR',
    start_factor=0.0001,
    end_factor=0.0000001,
    begin=50,
)

default_hooks = dict(
    checkpoint=dict(
        type='CheckpointHook',
        interval=1000,
        save_optimizer=True,
        by_epoch=False,
        out_dir=save_dir,
        save_best='auto',
    ),
    timer=dict(type='IterTimerHook'),
    logger=dict(type='LoggerHook', interval=1),
    param_scheduler=dict(type='ParamSchedulerHook'),
    sampler_seed=dict(type='DistSamplerSeedHook'),
)

load_from = 'https://download.openxlab.org.cn/models/xiaomile/DeblurGANv2/'\
            'weight/DeblurGANv2_fpn-mobilenet.pth'



_base_ = '../_base_/gen_default_runtime.py'

# config for model
dtype = 'fp16'
stable_diffusion_v15_url = 'runwayml/stable-diffusion-v1-5'

placeholder_token = '<cat-toy>'
initialize_token = 'toy'
num_vectors_per_token = 1
val_prompts = [
    'a <cat-toy> on packbag', 'a <cat-toy> on sofa',
    'a <cat-toy> in swimming pool', 'a <cat-toy>'
]

model = dict(
    type='TextualInversion',
    placeholder_token=placeholder_token,
    vae=dict(
        type='AutoencoderKL',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='vae'),
    unet=dict(
        type='UNet2DConditionModel',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='unet'),
    text_encoder=dict(
        type='ClipWrapper',
        clip_type='huggingface',
        pretrained_model_name_or_path=stable_diffusion_v15_url,
        subfolder='text_encoder'),
    tokenizer=stable_diffusion_v15_url,
    initialize_token=initialize_token,
    num_vectors_per_token=num_vectors_per_token,
    val_prompts=val_prompts,
    scheduler=dict(
        type='DDPMScheduler',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='scheduler'),
    test_scheduler=dict(
        type='DDIMScheduler',
        from_pretrained=stable_diffusion_v15_url,
        subfolder='scheduler'),
    data_preprocessor=dict(type='DataPreprocessor', data_keys=None))

train_cfg = dict(max_iters=3000)

optim_wrapper = dict(
    modules='.*trainable_embeddings',
    optimizer=dict(type='AdamW', lr=5e-4),
    accumulative_counts=1)

pipeline = [
    dict(type='LoadImageFromFile', key='img', channel_order='rgb'),
    dict(type='Resize', scale=(512, 512)),
    dict(type='PackInputs')
]

dataset = dict(
    type='TextualInversionDataset',
    data_root='./data/',
    concept_dir='cat_toy',
    placeholder=placeholder_token,
    template='data/imagenet_templates_small.txt',
    pipeline=pipeline)

train_dataloader = dict(
    dataset=dataset,
    num_workers=16,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    persistent_workers=True,
    batch_size=1)
val_cfg = val_evaluator = val_dataloader = None
test_cfg = test_evaluator = test_dataloader = None

default_hooks = dict(
    logger=dict(interval=10),
    checkpoint=dict(type='CheckpointHook', interval=10))
custom_hooks = [
    dict(
        type='VisualizationHook',
        interval=50,
        fixed_input=True,
        # visualize train dataset
        vis_kwargs_list=dict(type='Data', name='fake_img'),
        n_samples=1)
]



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
from mmengine.model import BaseModule

from mmagic.registry import MODELS
from ..basicvsr.basicvsr_net import BasicVSRNet, ResidualBlocksWithInputConv


@MODELS.register_module()
class RealBasicVSRNet(BaseModule):
    """RealBasicVSR network structure for real-world video super-resolution.

    Support only x4 upsampling.

    Paper:
        Investigating Tradeoffs in Real-World Video Super-Resolution, arXiv

    Args:
        mid_channels (int, optional): Channel number of the intermediate
            features. Default: 64.
        num_propagation_blocks (int, optional): Number of residual blocks in
            each propagation branch. Default: 20.
        num_cleaning_blocks (int, optional): Number of residual blocks in the
            image cleaning module. Default: 20.
        dynamic_refine_thres (int, optional): Stop cleaning the images when
            the residue is smaller than this value. Default: 255.
        spynet_pretrained (str, optional): Pre-trained model path of SPyNet.
            Default: None.
        is_fix_cleaning (bool, optional): Whether to fix the weights of
            the image cleaning module during training. Default: False.
        is_sequential_cleaning (bool, optional): Whether to clean the images
            sequentially. This is used to save GPU memory, but the speed is
            slightly slower. Default: False.
    """

    def __init__(self,
                 mid_channels=64,
                 num_propagation_blocks=20,
                 num_cleaning_blocks=20,
                 dynamic_refine_thres=255,
                 spynet_pretrained=None,
                 is_fix_cleaning=False,
                 is_sequential_cleaning=False):

        super().__init__()

        self.dynamic_refine_thres = dynamic_refine_thres / 255.
        self.is_sequential_cleaning = is_sequential_cleaning

        # image cleaning module
        self.image_cleaning = nn.Sequential(
            ResidualBlocksWithInputConv(3, mid_channels, num_cleaning_blocks),
            nn.Conv2d(mid_channels, 3, 3, 1, 1, bias=True),
        )

        if is_fix_cleaning:  # keep the weights of the cleaning module fixed
            self.image_cleaning.requires_grad_(False)

        # BasicVSR
        self.basicvsr = BasicVSRNet(mid_channels, num_propagation_blocks,
                                    spynet_pretrained)
        self.basicvsr.spynet.requires_grad_(False)

    def forward(self, lqs, return_lqs=False):
        """Forward function for BasicVSR++.

        Args:
            lqs (tensor): Input low quality (LQ) sequence with
                shape (n, t, c, h, w).
            return_lqs (bool): Whether to return LQ sequence. Default: False.

        Returns:
            Tensor: Output HR sequence.
        """
        n, t, c, h, w = lqs.size()

        for _ in range(0, 3):  # at most 3 cleaning, determined empirically
            if self.is_sequential_cleaning:
                residues = []
                for i in range(0, t):
                    residue_i = self.image_cleaning(lqs[:, i, :, :, :])
                    lqs[:, i, :, :, :] += residue_i
                    residues.append(residue_i)
                residues = torch.stack(residues, dim=1)
            else:  # time -> batch, then apply cleaning at once
                lqs = lqs.view(-1, c, h, w)
                residues = self.image_cleaning(lqs)
                lqs = (lqs + residues).view(n, t, c, h, w)

            # determine whether to continue cleaning
            if torch.mean(torch.abs(residues)) < self.dynamic_refine_thres:
                break

        # Super-resolution (BasicVSR)
        outputs = self.basicvsr(lqs)

        if return_lqs:
            return outputs, lqs
        else:
            return outputs



# Copyright (c) OpenMMLab. All rights reserved.
from .animatediff import AnimateDiff, UNet3DConditionMotionModel
from .aotgan import AOTBlockNeck, AOTEncoderDecoder, AOTInpaintor
from .arcface import IDLossModel
from .basicvsr import BasicVSR, BasicVSRNet
from .basicvsr_plusplus_net import BasicVSRPlusPlusNet
from .biggan import BigGAN
from .cain import CAIN, CAINNet
from .controlnet import ControlStableDiffusion
from .cyclegan import CycleGAN
from .dcgan import DCGAN
from .ddpm import DenoisingUnet
from .deblurganv2 import (DeblurGanV2, DeblurGanV2Discriminator,
                          DeblurGanV2Generator)
from .deepfillv1 import (ContextualAttentionModule, ContextualAttentionNeck,
                         DeepFillDecoder, DeepFillEncoder, DeepFillRefiner,
                         DeepFillv1Discriminators, DeepFillv1Inpaintor)
from .deepfillv2 import DeepFillEncoderDecoder
from .dic import (DIC, DICNet, FeedbackBlock, FeedbackBlockCustom,
                  FeedbackBlockHeatmapAttention, LightCNN, MaxFeature)
from .dim import DIM
from .disco_diffusion import ClipWrapper, DiscoDiffusion
from .dreambooth import DreamBooth
from .edsr import EDSRNet
from .edvr import EDVR, EDVRNet
from .eg3d import EG3D
from .esrgan import ESRGAN, RRDBNet
from .fastcomposer import FastComposer
from .fba import FBADecoder, FBAResnetDilated
from .flavr import FLAVR, FLAVRNet
from .gca import GCA
from .ggan import GGAN
from .glean import GLEANStyleGANv2
from .global_local import (GLDecoder, GLDilationNeck, GLEncoder,
                           GLEncoderDecoder)
from .guided_diffusion import AblatedDiffusionModel
from .iconvsr import IconVSRNet
from .indexnet import (DepthwiseIndexBlock, HolisticIndexBlock,
                       IndexedUpsample, IndexNet, IndexNetDecoder,
                       IndexNetEncoder)
from .inst_colorization import InstColorization
from .liif import LIIF, MLPRefiner
from .lsgan import LSGAN
from .mspie import MSPIEStyleGAN2, PESinGAN
from .nafnet import NAFBaseline, NAFBaselineLocal, NAFNet, NAFNetLocal
from .pconv import (MaskConvModule, PartialConv2d, PConvDecoder, PConvEncoder,
                    PConvEncoderDecoder, PConvInpaintor)
from .pggan import ProgressiveGrowingGAN
from .pix2pix import Pix2Pix
from .plain import PlainDecoder, PlainRefiner
from .rdn import RDNNet
from .real_basicvsr import RealBasicVSR, RealBasicVSRNet
from .real_esrgan import RealESRGAN, UNetDiscriminatorWithSpectralNorm
from .restormer import Restormer
from .sagan import SAGAN
from .singan import SinGAN
from .srcnn import SRCNNNet
from .srgan import SRGAN, ModifiedVGG, MSRResNet
from .stable_diffusion import StableDiffusion, StableDiffusionInpaint
from .stable_diffusion_xl import StableDiffusionXL
from .stylegan1 import StyleGAN1
from .stylegan2 import StyleGAN2
from .stylegan3 import StyleGAN3, StyleGAN3Generator
from .swinir import SwinIRNet
from .tdan import TDAN, TDANNet
from .textual_inversion import TextualInversion
from .tof import TOFlowVFINet, TOFlowVSRNet, ToFResBlock
from .ttsr import LTE, TTSR, SearchTransformer, TTSRDiscriminator, TTSRNet
from .vico import ViCo
from .wgan_gp import WGANGP

__all__ = [
    'AOTEncoderDecoder', 'AOTBlockNeck', 'AOTInpaintor',
    'ContextualAttentionNeck', 'ContextualAttentionModule', 'CAIN', 'CAINNet',
    'DIM', 'DIC', 'DICNet', 'LightCNN', 'FeedbackBlock',
    'FeedbackBlockHeatmapAttention', 'FeedbackBlockCustom', 'MaxFeature',
    'FLAVR', 'FLAVRNet', 'ToFResBlock', 'TOFlowVFINet', 'TOFlowVSRNet',
    'DeepFillEncoder', 'DeepFillEncoderDecoder', 'DeepFillDecoder',
    'DeepFillRefiner', 'DeepFillv1Inpaintor', 'DeepFillv1Discriminators',
    'EDSRNet', 'ESRGAN', 'DepthwiseIndexBlock', 'HolisticIndexBlock',
    'IndexNet', 'IndexNetEncoder', 'IndexedUpsample', 'IndexNetDecoder', 'GCA',
    'GLEncoderDecoder', 'GLEncoder', 'GLDecoder', 'GLDilationNeck',
    'PartialConv2d', 'PConvEncoderDecoder', 'PConvEncoder', 'PConvDecoder',
    'PConvInpaintor', 'MaskConvModule', 'RRDBNet', 'SRCNNNet', 'RRDBNet',
    'RealESRGAN', 'UNetDiscriminatorWithSpectralNorm', 'EDVR', 'EDVRNet',
    'TDAN', 'TDANNet', 'BasicVSR', 'BasicVSRNet', 'BasicVSRPlusPlusNet',
    'IconVSRNet', 'RealBasicVSR', 'RealBasicVSRNet', 'SRGAN', 'MaxFeature',
    'ModifiedVGG', 'MSRResNet', 'RDNNet', 'LTE', 'TTSR', 'TTSRNet',
    'TTSRDiscriminator', 'TTSRNet', 'SearchTransformer', 'GLEANStyleGANv2',
    'LIIF', 'MLPRefiner', 'PlainRefiner', 'PlainDecoder', 'FBAResnetDilated',
    'FBADecoder', 'WGANGP', 'CycleGAN', 'SAGAN', 'LSGAN', 'GGAN', 'Pix2Pix',
    'StyleGAN1', 'StyleGAN2', 'StyleGAN3', 'BigGAN', 'DCGAN',
    'ProgressiveGrowingGAN', 'SinGAN', 'AblatedDiffusionModel',
    'DiscoDiffusion', 'IDLossModel', 'PESinGAN', 'MSPIEStyleGAN2',
    'StyleGAN3Generator', 'InstColorization', 'NAFBaseline',
    'NAFBaselineLocal', 'NAFNet', 'NAFNetLocal', 'DenoisingUnet',
    'ClipWrapper', 'EG3D', 'Restormer', 'SwinIRNet', 'StableDiffusion',
    'ControlStableDiffusion', 'DreamBooth', 'TextualInversion', 'DeblurGanV2',
    'DeblurGanV2Generator', 'DeblurGanV2Discriminator',
    'StableDiffusionInpaint', 'ViCo', 'FastComposer', 'AnimateDiff',
    'UNet3DConditionMotionModel', 'StableDiffusionXL'
]



# Copyright (c) OpenMMLab. All rights reserved.
import random
from copy import deepcopy
from typing import Dict, List, Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmengine import print_log

from mmagic.models.archs import set_lora
from mmagic.registry import MODELS
from mmagic.structures import DataSample
from mmagic.utils.typing import SampleList
from ..stable_diffusion.stable_diffusion import StableDiffusion

ModelType = Union[Dict, nn.Module]


@MODELS.register_module()
class DreamBooth(StableDiffusion):
    """Implementation of `DreamBooth with Stable Diffusion.

    <https://arxiv.org/abs/2208.12242>`_ (DreamBooth).

    Args:
        vae (Union[dict, nn.Module]): The config or module for VAE model.
        text_encoder (Union[dict, nn.Module]): The config or module for text
            encoder.
        tokenizer (str): The **name** for CLIP tokenizer.
        unet (Union[dict, nn.Module]): The config or module for Unet model.
        schedule (Union[dict, nn.Module]): The config or module for diffusion
            scheduler.
        test_scheduler (Union[dict, nn.Module], optional): The config or
            module for diffusion scheduler in test stage (`self.infer`). If not
            passed, will use the same scheduler as `schedule`. Defaults to
            None.
        lora_config (dict, optional): The config for LoRA finetuning. Defaults
            to None.
        val_prompts (Union[str, List[str]], optional): The prompts for
            validation. Defaults to None.
        class_prior_prompt (str, optional): The prompt for class prior loss.
        num_class_images (int, optional): The number of images for class prior.
            Defaults to 3.
        prior_loss_weight (float, optional): The weight for class prior loss.
            Defaults to 0.
        finetune_text_encoder (bool, optional): Whether to fine-tune text
            encoder. Defaults to False.
        dtype (str, optional): The dtype for the model. Defaults to 'fp16'.
        enable_xformers (bool, optional): Whether to use xformers.
            Defaults to True.
        noise_offset_weight (bool, optional): The weight of noise offset
            introduced in https://www.crosslabs.org/blog/diffusion-with-offset-noise  # noqa
            Defaults to 0.
        tomesd_cfg (dict, optional): The config for TOMESD. Please refers to
            https://github.com/dbolya/tomesd and
            https://github.com/open-mmlab/mmagic/blob/main/mmagic/models/utils/tome_utils.py for detail.  # noqa
            Defaults to None.
        data_preprocessor (dict, optional): The pre-process config of
            :class:`BaseDataPreprocessor`. Defaults to
                dict(type='DataPreprocessor').
        init_cfg (dict, optional): The weight initialized config for
            :class:`BaseModule`. Defaults to None/
    """

    def __init__(self,
                 vae: ModelType,
                 text_encoder: ModelType,
                 tokenizer: str,
                 unet: ModelType,
                 scheduler: ModelType,
                 test_scheduler: Optional[ModelType] = None,
                 lora_config: Optional[dict] = None,
                 val_prompts: Union[str, List[str]] = None,
                 class_prior_prompt: Optional[str] = None,
                 num_class_images: Optional[int] = 3,
                 prior_loss_weight: float = 0,
                 finetune_text_encoder: bool = False,
                 dtype: str = 'fp16',
                 enable_xformers: bool = True,
                 noise_offset_weight: float = 0,
                 tomesd_cfg: Optional[dict] = None,
                 data_preprocessor: Optional[ModelType] = dict(
                     type='DataPreprocessor'),
                 init_cfg: Optional[dict] = None):

        super().__init__(vae, text_encoder, tokenizer, unet, scheduler,
                         test_scheduler, dtype, enable_xformers,
                         noise_offset_weight, tomesd_cfg, data_preprocessor,
                         init_cfg)
        self.num_class_images = num_class_images
        self.class_prior_prompt = class_prior_prompt
        self.prior_loss_weight = prior_loss_weight
        self.class_images = []

        self.dtype = torch.float32
        if dtype == 'fp16':
            self.dtype = torch.float16
        elif dtype == 'bf16':
            self.dtype = torch.bfloat16
        else:
            assert dtype in [
                'fp32', None
            ], ('dtype must be one of \'fp32\', \'fp16\', \'bf16\' or None.')

        self.finetune_text_encoder = finetune_text_encoder
        self.val_prompts = val_prompts
        self.lora_config = deepcopy(lora_config)

        self.prepare_model()
        self.set_lora()

    @torch.no_grad()
    def generate_class_prior_images(self, num_batches=None):
        """Generate images for class prior loss.

        Args:
            num_batches (int): Number of batches to generate images.
                If not passed, all images will be generated in one
                forward. Defaults to None.
        """
        if self.prior_loss_weight == 0:
            return
        if self.class_images:
            return

        assert self.class_prior_prompt is not None, (
            '\'class_prior_prompt\' must be set when \'prior_loss_weight\' is '
            'larger than 0.')
        assert self.num_class_images is not None, (
            '\'num_class_images\' must be set when \'prior_loss_weight\' is '
            'larger than 0.')

        print_log(
            'Generating class prior images with prompt: '
            f'{self.class_prior_prompt}', 'current')
        num_batches = num_batches or self.num_class_images

        unet_dtype = next(self.unet.parameters()).dtype
        self.unet.to(self.dtype)
        for idx in range(0, self.num_class_images, num_batches):
            prompt = self.class_prior_prompt
            if self.num_class_images > 1:
                prompt += f' {idx + 1} of {self.num_class_images}'

            output = self.infer(prompt, return_type='tensor')
            samples = output['samples']
            self.class_images.append(samples.clamp(-1, 1))
        self.unet.to(unet_dtype)

    def prepare_model(self):
        """Prepare model for training.

        Move model to target dtype and disable gradient for some models.
        """
        self.vae.requires_grad_(False)
        print_log('Set VAE untrainable.', 'current')
        self.vae.to(self.dtype)
        print_log(f'Move VAE to {self.dtype}.', 'current')
        if not self.finetune_text_encoder or self.lora_config:
            self.text_encoder.requires_grad_(False)
            print_log('Set Text Encoder untrainable.', 'current')
            self.text_encoder.to(self.dtype)
            print_log(f'Move Text Encoder to {self.dtype}.', 'current')
        if self.lora_config:
            self.unet.requires_grad_(False)
            print_log('Set Unet untrainable.', 'current')

    def set_lora(self):
        """Set LORA for model."""
        if self.lora_config:
            set_lora(self.unet, self.lora_config)

    @torch.no_grad()
    def val_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Calls
        ``self.data_preprocessor`` and ``self.infer`` in order. Return the
        generated results which will be passed to evaluator or visualizer.

        Args:
            data (dict or tuple or list): Data sampled from dataset.

        Returns:
            SampleList: Generated image or image dict.
        """
        data = self.data_preprocessor(data)
        data_samples = data['data_samples']
        if self.val_prompts is None:
            prompt = data_samples.prompt
        else:
            prompt = self.val_prompts
            # construct a fake data_sample for destruct
            data_samples.split() * len(prompt)
            data_samples = DataSample.stack(data_samples.split() * len(prompt))

        output = self.infer(prompt, return_type='tensor')
        samples = output['samples']
        samples = self.data_preprocessor.destruct(samples, data_samples)

        out_data_sample = DataSample(fake_img=samples, prompt=prompt)
        data_sample_list = out_data_sample.split()
        return data_sample_list

    @torch.no_grad()
    def test_step(self, data: dict) -> SampleList:
        """Gets the generated image of given data. Calls
        ``self.data_preprocessor`` and ``self.infer`` in order. Return the
        generated results which will be passed to evaluator or visualizer.

        Args:
            data (dict or tuple or list): Data sampled from dataset.

        Returns:
            SampleList: Generated image or image dict.
        """
        if self.val_prompts is None:
            data = self.data_preprocessor(data)
            data_samples = data['data_samples']
            prompt = data_samples.prompt
        else:
            prompt = self.val_prompts
            # construct a fake data_sample for destruct
            data_samples = DataSample.stack(data['data_samples'] * len(prompt))

        output = self.infer(prompt, return_type='tensor')
        samples = output['samples']
        samples = self.data_preprocessor.destruct(samples, data_samples)

        out_data_sample = DataSample(fake_img=samples, prompt=prompt)
        data_sample_list = out_data_sample.split()
        return data_sample_list

    def train_step(self, data, optim_wrapper):

        data = self.data_preprocessor(data)
        inputs, data_samples = data['inputs'], data['data_samples']

        with optim_wrapper.optim_context(self.unet):
            image = inputs  # image for new concept
            prompt = data_samples.prompt
            num_batches = image.shape[0]

            if self.prior_loss_weight != 0:
                # image and prompt for prior preservation
                self.generate_class_prior_images(num_batches=num_batches)
                class_images_used = []
                for _ in range(num_batches):
                    idx = random.randint(0, len(self.class_images) - 1)
                    class_images_used.append(self.class_images[idx])

                image = torch.cat([image, *class_images_used], dim=0)
                prompt = prompt + [self.class_prior_prompt]

            image = image.to(self.dtype)
            latents = self.vae.encode(image).latent_dist.sample()
            latents = latents * self.vae.config.scaling_factor

            noise = torch.randn_like(latents)
            timesteps = torch.randint(
                0,
                self.scheduler.num_train_timesteps, (num_batches, ),
                device=self.device)
            timesteps = timesteps.long()

            noisy_latents = self.scheduler.add_noise(latents, noise, timesteps)

            input_ids = self.tokenizer(
                prompt,
                max_length=self.tokenizer.model_max_length,
                return_tensors='pt',
                padding='max_length',
                truncation=True)['input_ids'].to(self.device)

            encoder_hidden_states = self.text_encoder(input_ids)[0]

            if self.scheduler.config.prediction_type == 'epsilon':
                gt = noise
            elif self.scheduler.config.prediction_type == 'v_prediction':
                gt = self.scheduler.get_velocity(latents, noise, timesteps)
            else:
                raise ValueError('Unknown prediction type '
                                 f'{self.scheduler.config.prediction_type}')

            # NOTE: we train unet in fp32, convert to float manually
            model_output = self.unet(
                noisy_latents.float(),
                timesteps,
                encoder_hidden_states=encoder_hidden_states.float())
            model_pred = model_output['sample']

            loss_dict = dict()
            if self.prior_loss_weight != 0:
                model_pred, prior_pred = model_pred.split(2, dim=1)
                gt, prior_gt = gt.split(2, dim=1)
                # calculate loss in FP32
                dreambooth_loss = F.mse_loss(model_pred.float(), gt.float())
                prior_loss = F.mse_loss(prior_pred.float(), prior_gt.float())
                loss_dict['dreambooth_loss'] = dreambooth_loss
                loss_dict['prior_loss'] = prior_loss * self.prior_loss_weight

            else:
                # calculate loss in FP32
                dreambooth_loss = F.mse_loss(model_pred.float(), gt.float())
                loss_dict['dreambooth_loss'] = dreambooth_loss

            parsed_loss, log_vars = self.parse_losses(loss_dict)
            optim_wrapper.update_params(parsed_loss)

        return log_vars

    def forward(self,
                inputs: torch.Tensor,
                data_samples: Optional[list] = None,
                mode: str = 'tensor') -> Union[Dict[str, torch.Tensor], list]:
        """forward is not implemented now."""
        raise NotImplementedError(
            'Forward is not implemented now, please use infer.')



# Copyright (c) OpenMMLab. All rights reserved.
# Adapted from https://github.com/huggingface/diffusers/
# blob/main/src/diffusers/models/unet_2d_condition.py

import json
import os
from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.utils.checkpoint
from diffusers.configuration_utils import ConfigMixin, register_to_config
from diffusers.models.embeddings import TimestepEmbedding, Timesteps
from diffusers.models.modeling_utils import ModelMixin
from diffusers.utils import BaseOutput
from huggingface_hub import snapshot_download
from mmengine.logging import MMLogger
from mmengine.model import constant_init
from safetensors import safe_open

from mmagic.registry import MODELS
from .resnet_3d import InflatedConv3d, InflatedGroupNorm
from .unet_block import (CrossAttnDownBlock3D, CrossAttnUpBlock3D, DownBlock3D,
                         UNetMidBlock3DCrossAttn, UpBlock3D, get_down_block,
                         get_up_block)

logger = MMLogger.get_current_instance()


@dataclass
class UNet3DConditionOutput(BaseOutput):
    """Output of UNet3DCondtion."""
    sample: torch.FloatTensor


@MODELS.register_module()
class UNet3DConditionMotionModel(ModelMixin, ConfigMixin):
    _supports_gradient_checkpointing = True
    """ Implementation of UNet3DConditionMotionModel"""

    @register_to_config
    def __init__(
        self,
        sample_size: Optional[int] = None,
        in_channels: int = 4,
        out_channels: int = 4,
        center_input_sample: bool = False,
        flip_sin_to_cos: bool = True,
        freq_shift: int = 0,
        down_block_types: Tuple[str] = (
            'CrossAttnDownBlock3D',
            'CrossAttnDownBlock3D',
            'CrossAttnDownBlock3D',
            'DownBlock3D',
        ),
        mid_block_type: str = 'UNetMidBlock3DCrossAttn',
        up_block_types: Tuple[str] = ('UpBlock3D', 'CrossAttnUpBlock3D',
                                      'CrossAttnUpBlock3D',
                                      'CrossAttnUpBlock3D'),
        only_cross_attention: Union[bool, Tuple[bool]] = False,
        block_out_channels: Tuple[int] = (320, 640, 1280, 1280),
        layers_per_block: int = 2,
        downsample_padding: int = 1,
        mid_block_scale_factor: float = 1,
        act_fn: str = 'silu',
        norm_num_groups: int = 32,
        norm_eps: float = 1e-5,
        cross_attention_dim: int = 768,
        attention_head_dim: Union[int, Tuple[int]] = 8,
        dual_cross_attention: bool = False,
        use_linear_projection: bool = False,
        class_embed_type: Optional[str] = None,
        num_class_embeds: Optional[int] = None,
        upcast_attention: bool = False,
        resnet_time_scale_shift: str = 'default',

        # Additional
        use_inflated_groupnorm=False,
        use_motion_module=False,
        motion_module_resolutions=(1, 2, 4, 8),
        motion_module_mid_block=False,
        motion_module_decoder_only=False,
        motion_module_type=None,
        motion_module_kwargs={},
        unet_use_cross_frame_attention=None,
        unet_use_temporal_attention=None,
        subfolder=None,
        from_pretrained=None,
        unet_addtion_kwargs=None,
    ):
        super().__init__()

        self.sample_size = sample_size
        time_embed_dim = block_out_channels[0] * 4

        # input
        self.conv_in = InflatedConv3d(
            in_channels, block_out_channels[0], kernel_size=3, padding=(1, 1))

        # time
        self.time_proj = Timesteps(block_out_channels[0], flip_sin_to_cos,
                                   freq_shift)
        timestep_input_dim = block_out_channels[0]

        self.time_embedding = TimestepEmbedding(timestep_input_dim,
                                                time_embed_dim)

        # class embedding
        if class_embed_type is None and num_class_embeds is not None:
            self.class_embedding = nn.Embedding(num_class_embeds,
                                                time_embed_dim)
        elif class_embed_type == 'timestep':
            self.class_embedding = TimestepEmbedding(timestep_input_dim,
                                                     time_embed_dim)
        elif class_embed_type == 'identity':
            self.class_embedding = nn.Identity(time_embed_dim, time_embed_dim)
        else:
            self.class_embedding = None

        self.down_blocks = nn.ModuleList([])
        self.mid_block = None
        self.up_blocks = nn.ModuleList([])

        if isinstance(only_cross_attention, bool):
            only_cross_attention = [only_cross_attention
                                    ] * len(down_block_types)

        if isinstance(attention_head_dim, int):
            attention_head_dim = (attention_head_dim, ) * len(down_block_types)

        # down
        output_channel = block_out_channels[0]
        for i, down_block_type in enumerate(down_block_types):
            res = 2**i
            input_channel = output_channel
            output_channel = block_out_channels[i]
            is_final_block = i == len(block_out_channels) - 1

            down_block = get_down_block(
                down_block_type,
                num_layers=layers_per_block,
                in_channels=input_channel,
                out_channels=output_channel,
                temb_channels=time_embed_dim,
                add_downsample=not is_final_block,
                resnet_eps=norm_eps,
                resnet_act_fn=act_fn,
                resnet_groups=norm_num_groups,
                cross_attention_dim=cross_attention_dim,
                attn_num_head_channels=attention_head_dim[i],
                downsample_padding=downsample_padding,
                dual_cross_attention=dual_cross_attention,
                use_linear_projection=use_linear_projection,
                only_cross_attention=only_cross_attention[i],
                upcast_attention=upcast_attention,
                resnet_time_scale_shift=resnet_time_scale_shift,
                unet_use_cross_frame_attention=unet_use_cross_frame_attention,
                unet_use_temporal_attention=unet_use_temporal_attention,
                use_inflated_groupnorm=use_inflated_groupnorm,
                use_motion_module=use_motion_module
                and (res in motion_module_resolutions)
                and (not motion_module_decoder_only),
                motion_module_type=motion_module_type,
                motion_module_kwargs=motion_module_kwargs,
            )
            self.down_blocks.append(down_block)

        # mid
        if mid_block_type == 'UNetMidBlock3DCrossAttn':
            self.mid_block = UNetMidBlock3DCrossAttn(
                in_channels=block_out_channels[-1],
                temb_channels=time_embed_dim,
                resnet_eps=norm_eps,
                resnet_act_fn=act_fn,
                output_scale_factor=mid_block_scale_factor,
                resnet_time_scale_shift=resnet_time_scale_shift,
                cross_attention_dim=cross_attention_dim,
                attn_num_head_channels=attention_head_dim[-1],
                resnet_groups=norm_num_groups,
                dual_cross_attention=dual_cross_attention,
                use_linear_projection=use_linear_projection,
                upcast_attention=upcast_attention,
                unet_use_cross_frame_attention=unet_use_cross_frame_attention,
                unet_use_temporal_attention=unet_use_temporal_attention,
                use_inflated_groupnorm=use_inflated_groupnorm,
                use_motion_module=use_motion_module
                and motion_module_mid_block,
                motion_module_type=motion_module_type,
                motion_module_kwargs=motion_module_kwargs,
            )
        else:
            raise ValueError(f'unknown mid_block_type : {mid_block_type}')

        # count how many layers upsample the videos
        self.num_upsamplers = 0

        # up
        reversed_block_out_channels = list(reversed(block_out_channels))
        reversed_attention_head_dim = list(reversed(attention_head_dim))
        only_cross_attention = list(reversed(only_cross_attention))
        output_channel = reversed_block_out_channels[0]
        for i, up_block_type in enumerate(up_block_types):
            res = 2**(3 - i)
            is_final_block = i == len(block_out_channels) - 1

            prev_output_channel = output_channel
            output_channel = reversed_block_out_channels[i]
            input_channel = reversed_block_out_channels[min(
                i + 1,
                len(block_out_channels) - 1)]

            # add upsample block for all BUT final layer
            if not is_final_block:
                add_upsample = True
                self.num_upsamplers += 1
            else:
                add_upsample = False

            up_block = get_up_block(
                up_block_type,
                num_layers=layers_per_block + 1,
                in_channels=input_channel,
                out_channels=output_channel,
                prev_output_channel=prev_output_channel,
                temb_channels=time_embed_dim,
                add_upsample=add_upsample,
                resnet_eps=norm_eps,
                resnet_act_fn=act_fn,
                resnet_groups=norm_num_groups,
                cross_attention_dim=cross_attention_dim,
                attn_num_head_channels=reversed_attention_head_dim[i],
                dual_cross_attention=dual_cross_attention,
                use_linear_projection=use_linear_projection,
                only_cross_attention=only_cross_attention[i],
                upcast_attention=upcast_attention,
                resnet_time_scale_shift=resnet_time_scale_shift,
                unet_use_cross_frame_attention=unet_use_cross_frame_attention,
                unet_use_temporal_attention=unet_use_temporal_attention,
                use_inflated_groupnorm=use_inflated_groupnorm,
                use_motion_module=use_motion_module
                and (res in motion_module_resolutions),
                motion_module_type=motion_module_type,
                motion_module_kwargs=motion_module_kwargs,
            )
            self.up_blocks.append(up_block)
            prev_output_channel = output_channel

        # out
        if use_inflated_groupnorm:
            self.conv_norm_out = InflatedGroupNorm(
                num_channels=block_out_channels[0],
                num_groups=norm_num_groups,
                eps=norm_eps)
        else:
            self.conv_norm_out = nn.GroupNorm(
                num_channels=block_out_channels[0],
                num_groups=norm_num_groups,
                eps=norm_eps)
        self.conv_norm_out = nn.GroupNorm(
            num_channels=block_out_channels[0],
            num_groups=norm_num_groups,
            eps=norm_eps)
        self.conv_act = nn.SiLU()
        self.conv_out = InflatedConv3d(
            block_out_channels[0], out_channels, kernel_size=3, padding=1)
        self.init_weights(subfolder, from_pretrained)

    def init_weights(self, subfolder=None, from_pretrained=None):
        """Init weights for models.

        We just use the initialization method proposed in the original paper.

        Args:
            pretrained (str, optional): Path for pretrained weights. If given
                None, pretrained weights will not be loaded. Defaults to None.
        """
        if isinstance(from_pretrained, str):
            from diffusers.utils import WEIGHTS_NAME
            model_file = os.path.join(from_pretrained, subfolder, WEIGHTS_NAME)
            if not os.path.isfile(model_file):
                cache_file = snapshot_download(
                    'runwayml/stable-diffusion-v1-5',
                    allow_patterns=['*.json', '*unet*safetensors'],
                    ignore_patterns=[
                        '*.fp16.safetensors', '*v1-5*', '*ema.safetensors'
                    ])
                from diffusers.utils import SAFETENSORS_WEIGHTS_NAME
                model_file = os.path.join(cache_file, subfolder,
                                          SAFETENSORS_WEIGHTS_NAME)
                state_dict = {}
                with safe_open(model_file, framework='pt', device='cpu') as f:
                    for key in f.keys():
                        state_dict[key] = f.get_tensor(key)
            else:
                state_dict = torch.load(model_file, map_location='cpu')

            m, u = self.load_state_dict(state_dict, strict=False)
            logger.info(
                f'### missing keys: {len(m)}; \n### unexpected keys: {len(u)};'
            )
            params = [
                p.numel() if 'temporal' in n else 0
                for n, p in self.named_parameters()
            ]
            logger.info(
                f'### Temporal Module Parameters: {sum(params) / 1e6} M')
        elif from_pretrained is None:
            #   As Improved-DDPM, we apply zero-initialization to
            #   second conv block in ResBlock (keywords: conv_2)
            #   the output layer of the Unet (keywords: 'out' but
            #     not 'out_blocks')
            #   projection layer in Attention layer (keywords: proj)
            for n, m in self.named_modules():
                if isinstance(m, nn.Conv2d) and ('conv2' in n or
                                                 ('out' in n
                                                  and 'out_blocks' not in n)):
                    constant_init(m, 0)
                if isinstance(m, nn.Conv1d) and 'proj' in n:
                    constant_init(m, 0)
        else:
            raise TypeError('from_pretrained must be a str or None but'
                            f' got {type(from_pretrained)} instead.')

    def set_attention_slice(self, slice_size):
        r"""
        Enable sliced attention computation.

        When this option is enabled, the attention module will
        split the input tensor in slices, to compute attention
        in several steps. This is useful to save some memory
        in exchange for a small speed decrease.

        Args:
            slice_size (`str` or `int` or `list(int)`, *optional*,
                defaults to `"auto"`):
                When `"auto"`, halves the input to the attention heads,
                so attention will be computed in two steps. If
                `"max"`, maximum amount of memory will be saved by
                running only one slice at a time. If a number is
                provided, uses as many slices as
                `attention_head_dim // slice_size`. In this case,
                `attention_head_dim' must be a multiple of `slice_size`.
        """
        sliceable_head_dims = []

        def fn_recursive_retrieve_slicable_dims(module: torch.nn.Module):
            """set attention slice recursively."""
            if hasattr(module, 'set_attention_slice'):
                sliceable_head_dims.append(module.sliceable_head_dim)

            for child in module.children():
                fn_recursive_retrieve_slicable_dims(child)

        # retrieve number of attention layers
        for module in self.children():
            fn_recursive_retrieve_slicable_dims(module)

        num_slicable_layers = len(sliceable_head_dims)

        if slice_size == 'auto':
            # half the attention head size is usually a good trade-off between
            # speed and memory
            slice_size = [dim // 2 for dim in sliceable_head_dims]
        elif slice_size == 'max':
            # make smallest slice possible
            slice_size = num_slicable_layers * [1]

        slice_size = num_slicable_layers * [slice_size] if not isinstance(
            slice_size, list) else slice_size

        if len(slice_size) != len(sliceable_head_dims):
            raise ValueError(
                f'You have provided {len(slice_size)}, but '
                f'{self.config} has {len(sliceable_head_dims)} different'
                f' attention layers. Make sure to match '
                f'`len(slice_size)` to be {len(sliceable_head_dims)}.')

        for i in range(len(slice_size)):
            size = slice_size[i]
            dim = sliceable_head_dims[i]
            if size is not None and size > dim:
                raise ValueError(
                    f'size {size} has to be smaller or equal to {dim}.')

        # Recursively walk through all the children.
        # Any children which exposes the set_attention_slice method
        # gets the message
        def fn_recursive_set_attention_slice(module: torch.nn.Module,
                                             slice_size: List[int]):
            """set attention slice recursively."""

            if hasattr(module, 'set_attention_slice'):
                module.set_attention_slice(slice_size.pop())

            for child in module.children():
                fn_recursive_set_attention_slice(child, slice_size)

        reversed_slice_size = list(reversed(slice_size))
        for module in self.children():
            fn_recursive_set_attention_slice(module, reversed_slice_size)

    def _set_gradient_checkpointing(self, module, value=False):
        """set gradient checkpoint."""
        if isinstance(module, (CrossAttnDownBlock3D, DownBlock3D,
                               CrossAttnUpBlock3D, UpBlock3D)):
            module.gradient_checkpointing = value

    def forward(
        self,
        sample: torch.FloatTensor,
        timestep: Union[torch.Tensor, float, int],
        encoder_hidden_states: torch.Tensor,
        class_labels: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        return_dict: bool = True,
    ) -> Union[UNet3DConditionOutput, Tuple]:
        r"""
        Args:
            sample (`torch.FloatTensor`): (batch, channel, height, width)
            noisy inputs tensor
            timestep (`torch.FloatTensor` or `float` or `int`):
            (batch) timesteps
            encoder_hidden_states (`torch.FloatTensor`):
            (batch, sequence_length, feature_dim) encoder hidden states
            return_dict (`bool`, *optional*, defaults to `True`):
                Whether or not to return a
                [`UNet3DConditionOutput`]
                instead of a plain tuple.

        Returns:
            [`UNet3DConditionOutput`] or `tuple`:
            [`UNet3DConditionOutput`]
            if `return_dict` is True, otherwise a `tuple`. When
            returning a tuple, the first element is the sample tensor.
        """
        # By default samples have to be AT least a multiple of the
        # overall upsampling factor. he overall upsampling factor is equal
        # T to 2 ** (# num of upsampling layears).
        # However, the upsampling interpolation output size
        # can be forced to fit any upsampling size
        # on the fly if necessary.
        default_overall_up_factor = 2**self.num_upsamplers

        # upsample size should be forwarded when sample is
        # not a multiple of `default_overall_up_factor`
        forward_upsample_size = False
        upsample_size = None

        if any(s % default_overall_up_factor != 0 for s in sample.shape[-2:]):
            logger.info(
                'Forward upsample size to force interpolation output size.')
            forward_upsample_size = True

        # prepare attention_mask
        if attention_mask is not None:
            attention_mask = (1 - attention_mask.to(sample.dtype)) * -10000.0
            attention_mask = attention_mask.unsqueeze(1)

        # center input if necessary
        if self.config.center_input_sample:
            sample = 2 * sample - 1.0

        # time
        timesteps = timestep
        if not torch.is_tensor(timesteps):
            # This would be a good case for the `match`
            # statement (Python 3.10+)
            is_mps = sample.device.type == 'mps'
            if isinstance(timestep, float):
                dtype = torch.float32 if is_mps else torch.float64
            else:
                dtype = torch.int32 if is_mps else torch.int64
            timesteps = torch.tensor([timesteps],
                                     dtype=dtype,
                                     device=sample.device)
        elif len(timesteps.shape) == 0:
            timesteps = timesteps[None].to(sample.device)

        # broadcast to batch dimension in a way
        # that's compatible with ONNX/Core ML
        timesteps = timesteps.expand(sample.shape[0])

        t_emb = self.time_proj(timesteps)

        # timesteps does not contain any weights and will
        # always return f32 tensors
        # but time_embedding might actually be running in fp16.
        # so we need to cast here.
        # there might be better ways to encapsulate this.
        t_emb = t_emb.to(dtype=self.dtype)
        emb = self.time_embedding(t_emb)

        if self.class_embedding is not None:
            if class_labels is None:
                raise ValueError(
                    'class_labels should be provided when num_class_embeds > 0'
                )

            if self.config.class_embed_type == 'timestep':
                class_labels = self.time_proj(class_labels)

            class_emb = self.class_embedding(class_labels).to(dtype=self.dtype)
            emb = emb + class_emb

        # pre-process
        sample = self.conv_in(sample)

        # down
        down_block_res_samples = (sample, )
        for downsample_block in self.down_blocks:
            if hasattr(downsample_block, 'has_cross_attention'
                       ) and downsample_block.has_cross_attention:
                sample, res_samples = downsample_block(
                    hidden_states=sample,
                    temb=emb,
                    encoder_hidden_states=encoder_hidden_states,
                    attention_mask=attention_mask,
                )
            else:
                sample, res_samples = downsample_block(
                    hidden_states=sample,
                    temb=emb,
                    encoder_hidden_states=encoder_hidden_states)

            down_block_res_samples += res_samples

        # mid
        sample = self.mid_block(
            sample,
            emb,
            encoder_hidden_states=encoder_hidden_states,
            attention_mask=attention_mask)

        # up
        for i, upsample_block in enumerate(self.up_blocks):
            is_final_block = i == len(self.up_blocks) - 1

            res_samples = down_block_res_samples[-len(upsample_block.resnets):]
            down_block_res_samples = down_block_res_samples[:-len(
                upsample_block.resnets)]

            # if we have not reached the final block and need to forward the
            # upsample size, we do it here
            if not is_final_block and forward_upsample_size:
                upsample_size = down_block_res_samples[-1].shape[2:]

            if hasattr(upsample_block, 'has_cross_attention'
                       ) and upsample_block.has_cross_attention:
                sample = upsample_block(
                    hidden_states=sample,
                    temb=emb,
                    res_hidden_states_tuple=res_samples,
                    encoder_hidden_states=encoder_hidden_states,
                    upsample_size=upsample_size,
                    attention_mask=attention_mask,
                )
            else:
                sample = upsample_block(
                    hidden_states=sample,
                    temb=emb,
                    res_hidden_states_tuple=res_samples,
                    upsample_size=upsample_size,
                    encoder_hidden_states=encoder_hidden_states,
                )

        # post-process
        sample = self.conv_norm_out(sample)
        sample = self.conv_act(sample)
        sample = self.conv_out(sample)

        if not return_dict:
            return (sample, )

        return UNet3DConditionOutput(sample=sample)

    @classmethod
    def from_pretrained_2d(cls,
                           pretrained_model_path,
                           subfolder=None,
                           unet_additional_kwargs=None):
        """a class method for initialization."""
        if subfolder is not None:
            pretrained_model_path = os.path.join(pretrained_model_path,
                                                 subfolder)
        logger.info(f"loaded temporal unet's pretrained weights \
                from {pretrained_model_path} ...")

        config_file = os.path.join(pretrained_model_path, 'config.json')
        if not os.path.isfile(config_file):
            raise RuntimeError(f'{config_file} does not exist')
        with open(config_file, 'r') as f:
            config = json.load(f)
        config['_class_name'] = cls.__name__
        config['down_block_types'] = [
            'CrossAttnDownBlock3D', 'CrossAttnDownBlock3D',
            'CrossAttnDownBlock3D', 'DownBlock3D'
        ]
        config['up_block_types'] = [
            'UpBlock3D', 'CrossAttnUpBlock3D', 'CrossAttnUpBlock3D',
            'CrossAttnUpBlock3D'
        ]

        from diffusers.utils import WEIGHTS_NAME
        model = cls.from_config(config, **unet_additional_kwargs)
        model_file = os.path.join(pretrained_model_path, WEIGHTS_NAME)
        if not os.path.isfile(model_file):
            raise RuntimeError(f'{model_file} does not exist')
        state_dict = torch.load(model_file, map_location='cpu')

        m, u = model.load_state_dict(state_dict, strict=False)
        logger.info(
            f'### missing keys: {len(m)}; \n### unexpected keys: {len(u)};')
        # print(f"### missing keys:\n{m}\n### unexpected keys:\n{u}\n")

        params = [
            p.numel() if 'temporal' in n else 0
            for n, p in model.named_parameters()
        ]
        logger.info(f'### Temporal Module Parameters: {sum(params) / 1e6} M')

        return model



# Copyright (c) OpenMMLab. All rights reserved.
from mmagic.models.archs import ResNet
from mmagic.registry import MODELS


@MODELS.register_module()
class FBAResnetDilated(ResNet):
    """ResNet-based encoder for FBA image matting."""

    def forward(self, x):
        """Forward function.

        Args:
            x (Tensor): Input tensor with shape (N, C, H, W).

        Returns:
            Tensor: Output tensor.
        """
        # x: (merged_t, trimap_t, two_channel_trimap,merged)
        # t refers to transformed.
        two_channel_trimap = x[:, 9:11]
        merged = x[:, 11:14]
        x = x[:, 0:11, ...]
        conv_out = [x]
        if self.deep_stem:
            x = self.stem(x)
        else:
            x = self.conv1(x)
            x = self.norm1(x)
            x = self.activate(x)
        conv_out.append(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        conv_out.append(x)
        x = self.layer2(x)
        conv_out.append(x)
        x = self.layer3(x)
        conv_out.append(x)
        x = self.layer4(x)
        conv_out.append(x)
        return {
            'conv_out': conv_out,
            'merged': merged,
            'two_channel_trimap': two_channel_trimap
        }



_base_ = [
    '../_base_/datasets/unconditional_imgs_128x128.py',
    '../_base_/gen_default_runtime.py',
]

# MODEL
loss_config = dict(gp_norm_mode='HWC', gp_loss_weight=10)
model = dict(
    type='WGANGP',
    data_preprocessor=dict(type='DataPreprocessor'),
    generator=dict(type='WGANGPGenerator', noise_size=128, out_scale=128),
    discriminator=dict(
        type='WGANGPDiscriminator',
        in_channel=3,
        in_scale=128,
        conv_module_cfg=dict(
            conv_cfg=None,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=True,
            act_cfg=dict(type='LeakyReLU', negative_slope=0.2),
            norm_cfg=dict(type='GN'),
            order=('conv', 'norm', 'act'))),
    discriminator_steps=5,
    loss_config=loss_config)

# `batch_size` and `data_root` need to be set.
batch_size = 64
data_root = './data/celeba-cropped/cropped_images_aligned_png/'
train_dataloader = dict(
    batch_size=batch_size, dataset=dict(data_root=data_root))

val_dataloader = dict(batch_size=batch_size, dataset=dict(data_root=data_root))

test_dataloader = dict(
    batch_size=batch_size, dataset=dict(data_root=data_root))

train_cfg = dict(max_iters=160000)

optim_wrapper = dict(
    generator=dict(optimizer=dict(type='Adam', lr=0.0001, betas=(0.5, 0.9))),
    discriminator=dict(
        optimizer=dict(type='Adam', lr=0.0001, betas=(0.5, 0.9))))

# VIS_HOOK
custom_hooks = [
    dict(
        type='VisualizationHook',
        interval=5000,
        fixed_input=True,
        vis_kwargs_list=dict(type='GAN', name='fake_img'))
]

# METRICS
metrics = [
    dict(
        type='MS_SSIM', prefix='ms-ssim', fake_nums=10000,
        sample_model='orig'),
    dict(
        type='SWD',
        prefix='swd',
        fake_nums=16384,
        sample_model='orig',
        image_shape=(3, 128, 128))
]

# save multi best checkpoints
default_hooks = dict(checkpoint=dict(save_best='swd/avg'))

val_evaluator = dict(metrics=metrics)
test_evaluator = dict(metrics=metrics)



_base_ = ['./dim_stage1-v16_1xb1-1000k_comp1k.py']
save_dir = './work_dirs/'
experiment_name = 'dim_stage1-v16_1xb1-1000k_comp1k_online-merge'

# dataset settings
train_pipeline = [
    dict(type='LoadImageFromFile', key='alpha', color_type='grayscale'),
    dict(type='LoadImageFromFile', key='fg'),
    dict(type='LoadImageFromFile', key='bg'),
    dict(type='MergeFgAndBg'),
    dict(
        type='CropAroundUnknown',
        keys=['alpha', 'merged', 'fg', 'bg'],
        crop_sizes=[320, 480, 640]),
    dict(type='Flip', keys=['alpha', 'merged', 'fg', 'bg']),
    dict(
        type='Resize',
        keys=['alpha', 'merged', 'fg', 'bg'],
        scale=(320, 320),
        keep_ratio=False),
    dict(type='GenerateTrimap', kernel_size=(1, 30)),
    dict(type='PackInputs'),
]

train_dataloader = dict(dataset=dict(pipeline=train_pipeline))



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os
import os.path as osp
import re
import sys
from multiprocessing import Pool

import cv2
import mmcv
import mmengine
import numpy as np


def worker(path, opt):
    """Worker for each process.

    Args:
        path (str): Image path.
        opt (dict): Configuration dict. It contains:
            crop_size (int): Crop size.
            step (int): Step for overlapped sliding window.
            thresh_size (int): Threshold size. Patches whose size is smaller
                than thresh_size will be dropped.
            save_folder (str): Path to save folder.
            compression_level (int): for cv2.IMWRITE_PNG_COMPRESSION.

    Returns:
        process_info (str): Process information displayed in progress bar.
    """
    crop_size = opt['crop_size']
    step = opt['step']
    thresh_size = opt['thresh_size']
    sequence, img_name = re.split(r'[\\/]', path)[-2:]
    img_name, extension = osp.splitext(osp.basename(path))

    img = mmcv.imread(path, flag='unchanged')

    if img.ndim == 2 or img.ndim == 3:
        h, w = img.shape[:2]
    else:
        raise ValueError(f'Image ndim should be 2 or 3, but got {img.ndim}')

    h_space = np.arange(0, h - crop_size + 1, step)
    if h - (h_space[-1] + crop_size) > thresh_size:
        h_space = np.append(h_space, h - crop_size)
    w_space = np.arange(0, w - crop_size + 1, step)
    if w - (w_space[-1] + crop_size) > thresh_size:
        w_space = np.append(w_space, w - crop_size)

    index = 0
    for x in h_space:
        for y in w_space:
            index += 1
            cropped_img = img[x:x + crop_size, y:y + crop_size, ...]
            sub_folder = osp.join(opt['save_folder'],
                                  f'{sequence}_s{index:03d}')
            mmengine.mkdir_or_exist(sub_folder)
            cv2.imwrite(
                osp.join(sub_folder, f'{img_name}{extension}'), cropped_img,
                [cv2.IMWRITE_PNG_COMPRESSION, opt['compression_level']])
    process_info = f'Processing {img_name} ...'
    return process_info


def extract_subimages(opt):
    """Crop images to subimages.

    Args:
        opt (dict): Configuration dict. It contains:
            input_folder (str): Path to the input folder.
            save_folder (str): Path to save folder.
            n_thread (int): Thread number.
    """
    input_folder = opt['input_folder']
    save_folder = opt['save_folder']
    if not osp.exists(save_folder):
        os.makedirs(save_folder)
        print(f'mkdir {save_folder} ...')
    else:
        print(f'Folder {save_folder} already exists. Exit.')
        sys.exit(1)

    img_list = list(mmengine.scandir(input_folder, recursive=True))

    img_list = [osp.join(input_folder, v) for v in img_list]
    prog_bar = mmengine.ProgressBar(len(img_list))
    pool = Pool(opt['n_thread'])
    for path in img_list:
        pool.apply_async(
            worker, args=(path, opt), callback=lambda arg: prog_bar.update())
    pool.close()
    pool.join()
    print('All processes done.')


def main_extract_subimages(args):
    """A multi-thread tool to crop large images to sub-images for faster IO.

    It is used for REDS dataset.

    opt (dict): Configuration dict. It contains:
        n_thread (int): Thread number.
        compression_level (int):  CV_IMWRITE_PNG_COMPRESSION from 0 to 9.
            A higher value means a smaller size and longer compression time.
            Use 0 for faster CPU decompression. Default: 3, same in cv2.

        scales (list[int]): The downsampling factors corresponding to the
            LR folders you want to process.
            Default: [].
        input_folder (str): Path to the input folder.
        save_folder (str): Path to save folder.
        crop_size (int): Crop size.
        step (int): Step for overlapped sliding window.
        thresh_size (int): Threshold size. Patches whose size is lower
            than thresh_size will be dropped.

    Usage:
        For each folder, run this script.
        For example, if scales = [4], there are two folders to be processed:
            train_sharp
            train_sharp_bicubic/X4
        After process, each sub_folder should have the same number of
        subimages. You can also specify scales by modifying the argument
        'scales'. Remember to modify opt configurations according to your
        settings.
    """

    opt = {}
    opt['n_thread'] = args.n_thread
    opt['compression_level'] = args.compression_level

    # HR images
    opt['input_folder'] = osp.join(args.data_root, 'train_sharp')
    opt['save_folder'] = osp.join(args.data_root, 'train_sharp_sub')
    opt['crop_size'] = args.crop_size
    opt['step'] = args.step
    opt['thresh_size'] = args.thresh_size
    extract_subimages(opt)

    for scale in args.scales:
        opt['input_folder'] = osp.join(args.data_root,
                                       f'train_sharp_bicubic/X{scale}')
        opt['save_folder'] = osp.join(args.data_root,
                                      f'train_sharp_bicubic/X{scale}_sub')
        opt['crop_size'] = args.crop_size // scale
        opt['step'] = args.step // scale
        opt['thresh_size'] = args.thresh_size // scale
        extract_subimages(opt)


def parse_args():
    parser = argparse.ArgumentParser(
        description='Preprocess REDS datasets',
        epilog='You can first download REDS datasets using the script from:'
        'https://gist.github.com/SeungjunNah/b10d369b92840cb8dd2118dd4f41d643')
    parser.add_argument('--data-root', type=str, help='root path for REDS')
    parser.add_argument(
        '--scales', nargs='*', default=[], help='scale factor list')
    parser.add_argument(
        '--crop-size',
        type=int,
        nargs='?',
        default=480,
        help='cropped size for HR images')
    parser.add_argument(
        '--step',
        type=int,
        nargs='?',
        default=240,
        help='step size for HR images')
    parser.add_argument(
        '--thresh-size',
        type=int,
        nargs='?',
        default=0,
        help='threshold size for HR images')
    parser.add_argument(
        '--compression-level',
        type=int,
        nargs='?',
        default=3,
        help='compression level when save png images')
    parser.add_argument(
        '--n-thread',
        type=int,
        nargs='?',
        default=20,
        help='thread number when using multiprocessing')

    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()

    # extract subimages
    args.scales = [int(v) for v in args.scales]
    main_extract_subimages(args)



# Copyright (c) OpenMMLab. All rights reserved.
# yapf: disable
import gradio as gr
import numpy as np
from PIL import Image, ImageDraw


class EasyDict(dict):
    """Convenience class that behaves like a dict but allows access with the
    attribute syntax."""

    def __getattr__(self, name: str):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name: str, value) -> None:
        self[name] = value

    def __delattr__(self, name: str) -> None:
        del self[name]


class ImageMask(gr.components.Image):
    """
    Sets: source="canvas", tool="sketch"
    """

    is_template = True

    def __init__(self, **kwargs):
        super().__init__(
            source='upload', tool='sketch', interactive=False, **kwargs)

    def preprocess(self, x):
        if x is None:
            return x
        if self.tool == 'sketch' and self.source in ['upload', 'webcam'
                                                     ] and type(x) != dict:
            decode_image = gr.processing_utils.decode_base64_to_image(x)
            width, height = decode_image.size
            mask = np.ones((height, width, 4), dtype=np.uint8)
            mask[..., -1] = 255
            mask = self.postprocess(mask)
            x = {'image': x, 'mask': mask}
        return super().preprocess(x)


def get_valid_mask(mask: np.ndarray):
    """Convert mask from gr.Image(0 to 255, RGBA) to binary mask."""
    if mask.ndim == 3:
        mask_pil = Image.fromarray(mask).convert('L')
        mask = np.array(mask_pil)
    if mask.max() == 255:
        mask = mask / 255
    return mask


def draw_points_on_image(image,
                         points,
                         curr_point=None,
                         highlight_all=True,
                         radius_scale=0.01):
    overlay_rgba = Image.new('RGBA', image.size, 0)
    overlay_draw = ImageDraw.Draw(overlay_rgba)
    for point_key, point in points.items():
        if ((curr_point is not None and curr_point == point_key)
                or highlight_all):
            p_color = (255, 0, 0)
            t_color = (0, 0, 255)

        else:
            p_color = (255, 0, 0, 35)
            t_color = (0, 0, 255, 35)

        rad_draw = int(image.size[0] * radius_scale)

        p_start = point.get('start_temp', point['start'])
        p_target = point['target']

        if p_start is not None and p_target is not None:
            p_draw = int(p_start[0]), int(p_start[1])
            t_draw = int(p_target[0]), int(p_target[1])

            overlay_draw.line(
                (p_draw[0], p_draw[1], t_draw[0], t_draw[1]),
                fill=(255, 255, 0),
                width=2,
            )

        if p_start is not None:
            p_draw = int(p_start[0]), int(p_start[1])
            overlay_draw.ellipse(
                (
                    p_draw[0] - rad_draw,
                    p_draw[1] - rad_draw,
                    p_draw[0] + rad_draw,
                    p_draw[1] + rad_draw,
                ),
                fill=p_color,
            )

            if curr_point is not None and curr_point == point_key:
                overlay_draw.text(p_draw, 'p', align='center', fill=(0, 0, 0))

        if p_target is not None:
            t_draw = int(p_target[0]), int(p_target[1])
            overlay_draw.ellipse(
                (
                    t_draw[0] - rad_draw,
                    t_draw[1] - rad_draw,
                    t_draw[0] + rad_draw,
                    t_draw[1] + rad_draw,
                ),
                fill=t_color,
            )

            if curr_point is not None and curr_point == point_key:
                overlay_draw.text(t_draw, 't', align='center', fill=(0, 0, 0))

    return Image.alpha_composite(image.convert('RGBA'),
                                 overlay_rgba).convert('RGB')


def draw_mask_on_image(image, mask):
    im_mask = np.uint8(mask * 255)
    im_mask_rgba = np.concatenate(
        (
            np.tile(im_mask[..., None], [1, 1, 3]),
            45 * np.ones(
                (im_mask.shape[0], im_mask.shape[1], 1), dtype=np.uint8),
        ),
        axis=-1,
    )
    im_mask_rgba = Image.fromarray(im_mask_rgba).convert('RGBA')

    return Image.alpha_composite(image.convert('RGBA'),
                                 im_mask_rgba).convert('RGB')


def on_change_single_global_state(keys,
                                  value,
                                  global_state,
                                  map_transform=None):
    if map_transform is not None:
        value = map_transform(value)

    curr_state = global_state
    if isinstance(keys, str):
        last_key = keys

    else:
        for k in keys[:-1]:
            curr_state = curr_state[k]

        last_key = keys[-1]

    curr_state[last_key] = value
    return global_state


def get_latest_points_pair(points_dict):
    if not points_dict:
        return None
    point_idx = list(points_dict.keys())
    latest_point_idx = max(point_idx)
    return latest_point_idx



# Copyright (c) OpenMMLab. All rights reserved.
import datetime
import os

import cv2
from component import ConcatImageWidget, QLabelSlider, VideoSlider
from PyQt5 import QtCore, QtWidgets
from utils import layout2widget


class PatchTab(QtWidgets.QWidget):

    def __init__(self, parent):
        super().__init__()

        self.parent = parent
        self.statusBar = self.parent.statusBar
        self.file_paths = []
        self.labels = []
        self.rect = None
        self.images = None
        self.isShow = False

        # Left Widget
        self.btn_add_file = QtWidgets.QPushButton()
        self.btn_add_file.setText('Add')
        self.btn_add_file.clicked.connect(self.add_file)
        self.btn_open_file = QtWidgets.QPushButton()
        self.btn_open_file.setText('Open new')
        self.btn_open_file.clicked.connect(self.open_file)

        self.input_label = QtWidgets.QLineEdit()
        self.input_file = QtWidgets.QLineEdit()

        self.tb_files = QtWidgets.QTableWidget()
        self.tb_files.setColumnCount(3)
        self.tb_files.setHorizontalHeaderLabels(
            ['File/Folder', 'Label', 'Other'])
        self.tb_files.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.Stretch)
        self.tb_files.setSelectionBehavior(
            QtWidgets.QAbstractItemView.SelectRows)
        self.tb_files.setSelectionMode(
            QtWidgets.QAbstractItemView.SingleSelection)
        self.tb_files.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.tb_files.customContextMenuRequested.connect(self.tableMenu)

        left_grid = QtWidgets.QGridLayout()
        left_grid.addWidget(self.tb_files, 0, 0, 5, 10)
        left_grid.addWidget(self.input_file, 6, 0, 1, 4)
        left_grid.addWidget(self.input_label, 6, 4, 1, 2)
        left_grid.addWidget(self.btn_open_file, 6, 6, 1, 2)
        left_grid.addWidget(self.btn_add_file, 6, 8, 1, 2)

        # Right Widget
        styleSheet = '''
            QGroupBox {
                background-color: qlineargradient(x1: 0, y1: 0, x2: 0, y2: 1,
                                                  stop: 0 #E0E0E0, stop: 1 #FFFFFF);
                border: 1px solid #999999;
                border-radius: 5px;
                margin-top: 2ex;  /*leave space at the top for the title */
                padding: 2ex 10ex;
                font-size: 20px;
                color: black;
            }
            QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;    /* position at the top center */
                padding: 0 3px;
                left: 30px;
                font-size: 8px;
                color: black;
            }
        ''' # noqa
        # select mode
        self.modeRect = QtWidgets.QGroupBox('Select Mode')
        self.modeRect.setFlat(True)
        # self.modeRect.setStyleSheet(styleSheet)
        btn_mode1 = QtWidgets.QRadioButton('Input whole image', self.modeRect)
        btn_mode2 = QtWidgets.QRadioButton('Input crop image', self.modeRect)
        self.btnGroup_mode = QtWidgets.QButtonGroup()
        self.btnGroup_mode.addButton(btn_mode1, 0)
        self.btnGroup_mode.addButton(btn_mode2, 1)
        self.btnGroup_mode.button(0).setChecked(True)
        self.btnGroup_mode.idToggled.connect(self.reset)
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.addWidget(btn_mode1)
        hlayout.addWidget(btn_mode2)
        self.modeRect.setLayout(hlayout)

        # select dirType
        self.dirTypeRect = QtWidgets.QGroupBox('Select File or Directory')
        self.dirTypeRect.setFlat(True)
        # self.dirTypeRect.setStyleSheet(styleSheet)
        btn_dirType1 = QtWidgets.QRadioButton('Directory', self.dirTypeRect)
        btn_dirType2 = QtWidgets.QRadioButton('File', self.dirTypeRect)
        self.btnGroup_dirType = QtWidgets.QButtonGroup()
        self.btnGroup_dirType.addButton(btn_dirType1, 0)
        self.btnGroup_dirType.addButton(btn_dirType2, 1)
        self.btnGroup_dirType.button(0).setChecked(True)
        self.btnGroup_dirType.idToggled.connect(self.reset)
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.addWidget(btn_dirType1)
        hlayout.addWidget(btn_dirType2)
        self.dirTypeRect.setLayout(hlayout)

        # select gt
        self.cb_gt = QtWidgets.QComboBox()
        self.cb_gt.currentTextChanged.connect(self.change_gt)

        # set column
        self.spin_cols = QtWidgets.QSpinBox()
        self.spin_cols.setMinimum(1)
        self.spin_cols.setMaximum(10)
        self.spin_cols.setValue(4)
        self.spin_cols.valueChanged.connect(self.set_column)

        # set scale
        self.scale = 100
        self.txt_scale = QtWidgets.QLabel('100 %')
        self.slider_scale = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider_scale.setMinimum(1)
        self.slider_scale.setMaximum(200)
        self.slider_scale.setValue(100)
        self.slider_scale.valueChanged.connect(self.set_scale)

        # operation
        self.btn_run = QtWidgets.QPushButton()
        self.btn_run.setText('Run')
        self.btn_run.clicked.connect(self.run)
        self.btn_reset = QtWidgets.QPushButton()
        self.btn_reset.setText('Reset')
        self.btn_reset.clicked.connect(self.reset)
        self.btn_save = QtWidgets.QPushButton()
        self.btn_save.setText('Save')
        self.btn_save.clicked.connect(self.save)

        right_grid = QtWidgets.QGridLayout()
        right_grid.addWidget(self.modeRect, 0, 0, 1, 20)
        right_grid.addWidget(self.dirTypeRect, 1, 0, 1, 20)
        right_grid.addWidget(self.cb_gt, 2, 0, 1, 15)
        right_grid.addWidget(QtWidgets.QLabel('Reference'), 2, 16, 1, 3)
        right_grid.addWidget(self.spin_cols, 3, 0, 1, 15)
        right_grid.addWidget(QtWidgets.QLabel('Set   columns'), 3, 16, 1, 3)
        right_grid.addWidget(self.txt_scale, 4, 0, 1, 1)
        right_grid.addWidget(self.slider_scale, 4, 1, 1, 14)
        right_grid.addWidget(QtWidgets.QLabel('Set   scale'), 4, 16, 1, 3)
        right_grid.addWidget(self.btn_run, 5, 0, 1, 20)
        right_grid.addWidget(self.btn_reset, 6, 0, 1, 20)
        right_grid.addWidget(self.btn_save, 7, 0, 1, 20)

        # Bottom Widget
        self.image_scroll = QtWidgets.QScrollArea()
        self.image_scroll.installEventFilter(self)

        # Splitter
        hsplitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        vsplitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        hsplitter.addWidget(layout2widget(left_grid))
        hsplitter.addWidget(layout2widget(right_grid))
        hsplitter.setStretchFactor(0, 1)
        hsplitter.setStretchFactor(1, 2)
        vsplitter.addWidget(hsplitter)
        vsplitter.addWidget(self.image_scroll)
        vsplitter.setStretchFactor(0, 1)
        vsplitter.setStretchFactor(1, 5)
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.addWidget(vsplitter)
        self.setLayout(hlayout)

    def open_file(self):
        """Open a file or directory from dialog."""
        if self.btnGroup_dirType.checkedId() == 0:
            path = QtWidgets.QFileDialog.getExistingDirectory()
            if len(path) <= 0:
                return
            label = path.split('/')[-1]
            self.input_file.setText(path)
            self.input_label.setText(label)
        else:
            path, _ = QtWidgets.QFileDialog.getOpenFileName(
                self, 'Select file', '', 'Images (*.jpg *.png)')
            if len(path) <= 0:
                return
            label = path.split('/')[-1].split('.')[0]
            self.input_file.setText(path)
            self.input_label.setText(label)

    def add_file(self):
        """Add opened file or directory to table."""
        if os.path.exists(self.input_file.text()):
            row = self.tb_files.rowCount()
            self.tb_files.setRowCount(row + 1)
            self.tb_files.setItem(
                row, 0, QtWidgets.QTableWidgetItem(self.input_file.text()))
            self.tb_files.setItem(
                row, 1, QtWidgets.QTableWidgetItem(self.input_label.text()))
            # if self.cb_gt.count() == 0:
            #     self.set_gt(self.input_file.text())
        else:
            QtWidgets.QMessageBox.about(self, 'Message',
                                        'Please input available file/folder')

    def tableMenu(self, pos):
        """Set mouse right button menu of table."""
        menu = QtWidgets.QMenu()
        menu_up = menu.addAction('move up')
        menu_down = menu.addAction('move down')
        menu_delete = menu.addAction('delete')
        menu_gt = menu.addAction('set reference')
        menu_pos = self.tb_files.mapToGlobal(pos)
        row = None
        for i in self.tb_files.selectionModel().selection().indexes():
            row = i.row()
        if row is None or row >= self.tb_files.rowCount():
            return
        action = menu.exec(menu_pos)
        if action == menu_up:
            self.tb_swap(row, -1)
        elif action == menu_down:
            self.tb_swap(row, 1)
        elif action == menu_delete:
            self.tb_files.removeRow(row)
        elif action == menu_gt:
            for r in range(self.tb_files.rowCount()):
                self.tb_files.setItem(r, 2, QtWidgets.QTableWidgetItem(''))
            self.tb_files.setItem(row, 2,
                                  QtWidgets.QTableWidgetItem('Reference'))
            self.set_gt(self.tb_files.item(row, 0).text())
        if self.isShow:
            self.run()

    def tb_swap(self, row, move):
        """Move items of table."""
        if move == -1:
            if 0 == row:
                return
            target = row - 1
        elif move == 1:
            if self.tb_files.rowCount() - 1 == row:
                return
            target = row + 1
        for col in range(self.tb_files.columnCount()):
            tmp = self.tb_files.takeItem(row, col)
            self.tb_files.setItem(row, col,
                                  self.tb_files.takeItem(target, col))
            self.tb_files.setItem(target, col, tmp)

    def set_gt(self, path):
        """Set GT ComboBox items."""
        self.cb_gt.clear()
        if os.path.isfile(path):
            self.cb_gt.addItem(path)
        else:
            files = sorted(os.listdir(path))
            for f in files:
                self.cb_gt.addItem(path + '/' + f)

    def change_gt(self):
        if self.isShow:
            self.run()

    def set_scale(self):
        """Set scale."""
        scale = self.slider_scale.value()
        self.txt_scale.setText(f'{scale} %')
        self.scale = scale
        if self.isShow:
            rect = None
            if self.rect:
                rect = [
                    int(r * scale / 100.0 / self.old_scale) for r in self.rect
                ]
            self.run(rect)

    def set_column(self):
        """Set column."""
        if self.isShow:
            self.run()

    def run(self, rect=None):
        """Generate patch compare result."""
        if self.cb_gt.currentText() == '':
            QtWidgets.QMessageBox.about(self, 'Message', 'Please set gt!')
            return

        rows = self.tb_files.rowCount()
        if rows <= 0:
            QtWidgets.QMessageBox.about(self, 'Message',
                                        'Please add a file at least!')
            return

        files = []
        self.labels = []
        for r in range(rows):
            if self.tb_files.item(r, 2).text() != 'Reference':
                files.append(self.tb_files.item(r, 0).text())
                self.labels.append(self.tb_files.item(r, 1).text())

        file_name = os.path.basename(self.cb_gt.currentText())
        self.file_paths = []
        for i, file in enumerate(files):
            if self.btnGroup_dirType.checkedId() == 0:
                self.file_paths.append(file + '/' + file_name)
            else:
                self.file_paths.append(file)

        mode = self.btnGroup_mode.checkedId()
        self.images = ConcatImageWidget(self, mode, self.spin_cols.value())
        self.images.set_images(self.file_paths, self.labels,
                               self.cb_gt.currentText(), self.scale / 100.0,
                               rect)
        self.image_scroll.setWidget(self.images)
        self.isShow = True

    def reset(self):
        """Init window."""
        self.file_paths = []
        self.labels = []
        self.isShow = False

        self.input_label.clear()
        self.input_file.clear()
        self.tb_files.setRowCount(0)
        self.cb_gt.clear()
        self.spin_cols.setValue(4)
        self.slider_scale.setValue(100)
        self.images = ConcatImageWidget(self, 0, self.spin_cols.value())
        self.image_scroll.setWidget(self.images)

    def save(self):
        """Save patch compare result."""
        path, _ = QtWidgets.QFileDialog.getSaveFileName(self, 'save')
        if len(path) <= 0:
            return
        if self.images:
            self.images.grab().save(path)
            QtWidgets.QMessageBox.about(self, 'Message', 'Success!')
        else:
            QtWidgets.QMessageBox.about(self, 'Message', 'Nothing to save.')

    def wheelEvent(self, ev) -> None:
        key = QtWidgets.QApplication.keyboardModifiers()
        if key == QtCore.Qt.ControlModifier:
            scale = ev.angleDelta().y() / 120
            self.scale += scale
            if self.scale < 1:
                self.scale = 1
            self.txt_scale.setText(f'{self.scale} %')
            self.slider_scale.setValue(self.scale)
            if self.scale > 200 and self.isShow:
                rect = None
                if self.rect:
                    rect = [
                        int(r * self.scale / 100.0 / self.old_scale)
                        for r in self.rect
                    ]
                self.run(rect)
        return super().wheelEvent(ev)

    def keyPressEvent(self, ev) -> None:
        if ev.key() == QtCore.Qt.Key_Left:
            if self.cb_gt.currentIndex() > 0:
                self.cb_gt.setCurrentIndex(self.cb_gt.currentIndex() - 1)
            else:
                self.cb_gt.setCurrentIndex(self.cb_gt.count() - 1)
        elif ev.key() == QtCore.Qt.Key_Right:
            if self.cb_gt.currentIndex() < self.cb_gt.count() - 1:
                self.cb_gt.setCurrentIndex(self.cb_gt.currentIndex() + 1)
            else:
                self.cb_gt.setCurrentIndex(0)
        return super().keyPressEvent(ev)

    def eventFilter(self, object, event) -> bool:
        if object == self.image_scroll:
            if event.type() == QtCore.QEvent.KeyPress:
                self.keyPressEvent(event)
                return False
        return super().eventFilter(object, event)


class SliderTab(QtWidgets.QWidget):

    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.images = [None, None]
        self.imageArea = None

        # Type setting
        self.typeRect = QtWidgets.QGroupBox('Type')
        self.typeRect.setFlat(True)
        btn_type1 = QtWidgets.QRadioButton('Video', self.typeRect)
        btn_type2 = QtWidgets.QRadioButton('Image', self.typeRect)
        self.btnGroup_type = QtWidgets.QButtonGroup()
        self.btnGroup_type.addButton(btn_type1, 0)
        self.btnGroup_type.addButton(btn_type2, 1)
        self.btnGroup_type.button(0).setChecked(True)
        self.btnGroup_type.idToggled.connect(self.reset)
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.addWidget(btn_type1)
        hlayout.addWidget(btn_type2)
        self.typeRect.setLayout(hlayout)

        # Mode setting
        self.modeRect = QtWidgets.QGroupBox('Mode')
        self.modeRect.setFlat(True)
        btn_mode1 = QtWidgets.QRadioButton('Match', self.modeRect)
        btn_mode2 = QtWidgets.QRadioButton('Single', self.modeRect)
        self.btnGroup_mode = QtWidgets.QButtonGroup()
        self.btnGroup_mode.addButton(btn_mode1, 0)
        self.btnGroup_mode.addButton(btn_mode2, 1)
        self.btnGroup_mode.button(0).setChecked(True)
        self.btnGroup_mode.idToggled.connect(self.reset)
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.addWidget(btn_mode1)
        hlayout.addWidget(btn_mode2)
        self.modeRect.setLayout(hlayout)

        # Settings
        self.cb_1 = QtWidgets.QComboBox()
        self.cb_2 = QtWidgets.QComboBox()
        self.cb_1.currentIndexChanged.connect(self.change_image_1)
        self.cb_2.currentIndexChanged.connect(self.change_image_2)
        self.input_label_1 = QtWidgets.QLineEdit()
        self.input_label_2 = QtWidgets.QLineEdit()
        self.input_title = QtWidgets.QLineEdit()
        self.input_label_1.textChanged.connect(self.set_label)
        self.input_label_2.textChanged.connect(self.set_label)
        self.input_title.textChanged.connect(self.set_label)

        # Set scale
        self.scale = 100
        self.txt_scale = QtWidgets.QLabel('100 %')
        self.slider_scale = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider_scale.setMinimum(1)
        self.slider_scale.setMaximum(200)
        self.slider_scale.setValue(100)
        self.slider_scale.valueChanged.connect(self.set_scale)

        # Auto slider setting
        self.autoRect = QtWidgets.QGroupBox('Auto Slider')
        self.autoRect.setFlat(True)
        btn_autoMode1 = QtWidgets.QRadioButton('Right', self.autoRect)
        btn_autoMode2 = QtWidgets.QRadioButton('Left', self.autoRect)
        btn_autoMode3 = QtWidgets.QRadioButton('Alternate', self.autoRect)
        self.btnGroup_auto = QtWidgets.QButtonGroup()
        self.btnGroup_auto.addButton(btn_autoMode1, 0)
        self.btnGroup_auto.addButton(btn_autoMode2, 1)
        self.btnGroup_auto.addButton(btn_autoMode3, 2)
        self.btnGroup_auto.button(2).setChecked(True)
        self.btnGroup_auto.idToggled.connect(self.set_autoSlider)
        self.slider_auto = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider_auto.setMinimum(0)
        self.slider_auto.setMaximum(200)
        self.slider_auto.setValue(0)
        self.slider_auto.valueChanged.connect(self.set_autoSlider)
        glayout = QtWidgets.QGridLayout()
        glayout.addWidget(QtWidgets.QLabel('Direction:'), 0, 0, 1, 1)
        glayout.addWidget(btn_autoMode3, 0, 2, 1, 1)
        glayout.addWidget(btn_autoMode1, 0, 3, 1, 1)
        glayout.addWidget(btn_autoMode2, 0, 4, 1, 1)
        glayout.addWidget(QtWidgets.QLabel('Speed:'), 1, 0, 1, 1)
        glayout.addWidget(self.slider_auto, 1, 1, 1, 4)
        self.autoRect.setLayout(glayout)

        # Add file
        self.btn_add_1 = QtWidgets.QPushButton()
        self.btn_add_2 = QtWidgets.QPushButton()
        self.btn_add_1.setText('Add a video')
        self.btn_add_2.setText('Add frames')
        self.btn_add_1.clicked.connect(self.add_1)
        self.btn_add_2.clicked.connect(self.add_2)

        # Buttons
        self.btn_pause = QtWidgets.QPushButton()
        self.btn_pause.setText('Pause (Space)')
        self.btn_pause.clicked.connect(self.pause)
        self.btn_pause.setEnabled(False)

        self.btn_reset = QtWidgets.QPushButton()
        self.btn_reset.setText('Reset')
        self.btn_reset.clicked.connect(self.reset)

        self.btn_save = QtWidgets.QPushButton()
        self.btn_save.setText('Save')
        self.btn_save.clicked.connect(self.save)

        self.btn_record = QtWidgets.QPushButton()
        self.btn_record.setText('Record (Enter)')
        self.btn_record.clicked.connect(self.record)

        self.txt_prompt = QtWidgets.QLabel()
        self.txt_prompt.setStyleSheet('color: red; font-size: 28px')

        left_grid = QtWidgets.QGridLayout()
        left_grid.addWidget(self.typeRect, 0, 0, 1, 10)
        left_grid.addWidget(self.modeRect, 1, 0, 1, 10)
        left_grid.addWidget(QtWidgets.QLabel('Left'), 2, 0, 1, 1)
        left_grid.addWidget(self.cb_1, 2, 1, 1, 9)
        left_grid.addWidget(QtWidgets.QLabel('Right'), 3, 0, 1, 1)
        left_grid.addWidget(self.cb_2, 3, 1, 1, 9)
        left_grid.addWidget(QtWidgets.QLabel('Set label 1'), 4, 0, 1, 1)
        left_grid.addWidget(self.input_label_1, 4, 1, 1, 9)
        left_grid.addWidget(QtWidgets.QLabel('Set label 2'), 5, 0, 1, 1)
        left_grid.addWidget(self.input_label_2, 5, 1, 1, 9)
        left_grid.addWidget(QtWidgets.QLabel('Set title'), 6, 0, 1, 1)
        left_grid.addWidget(self.input_title, 6, 1, 1, 9)
        left_grid.addWidget(QtWidgets.QLabel('Set scale'), 7, 0, 1, 1)
        left_grid.addWidget(self.slider_scale, 7, 1, 1, 8)
        left_grid.addWidget(self.txt_scale, 7, 9, 1, 1)
        left_grid.addWidget(self.autoRect, 8, 0, 1, 10)
        left_grid.addWidget(self.btn_add_1, 9, 0, 1, 5)
        left_grid.addWidget(self.btn_add_2, 9, 5, 1, 5)
        left_grid.addWidget(self.btn_pause, 10, 0, 1, 10)
        left_grid.addWidget(self.btn_reset, 11, 0, 1, 10)
        left_grid.addWidget(self.btn_save, 12, 0, 1, 10)
        left_grid.addWidget(self.btn_record, 13, 0, 1, 10)
        left_grid.addWidget(QtWidgets.QLabel(), 14, 0, 10, 10)
        left_grid.addWidget(self.txt_prompt, 25, 0, 20, 10)

        # Image area
        self.image_scroll = QtWidgets.QScrollArea()
        self.image_scroll.setAlignment(QtCore.Qt.AlignCenter)
        self.image_scroll.installEventFilter(self)
        right_grid = QtWidgets.QGridLayout()
        right_grid.addWidget(self.image_scroll, 0, 0)

        # Splitter
        hsplitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        hsplitter.addWidget(layout2widget(left_grid))
        hsplitter.addWidget(layout2widget(right_grid))
        hsplitter.setStretchFactor(0, 1)
        hsplitter.setStretchFactor(1, 5)
        hlayout = QtWidgets.QHBoxLayout()
        hlayout.addWidget(hsplitter)
        self.setLayout(hlayout)

        # Timer
        self.timer_slider = QtCore.QTimer(self)
        self.timer_slider.timeout.connect(self.auto_slider)
        self.timer_record = QtCore.QTimer(self)
        self.timer_record.timeout.connect(self.recording)

        # Player
        self.player = VideoSlider(self)
        self.player.sigout.connect(self.setImg)
        # self.player1_end = False
        # self.player2_end = False
        # self.player1 = VideoPlayer(self)
        # self.player2 = VideoPlayer(self)
        # self.player1.sigout.connect(self.setImg1)
        # self.player2.sigout.connect(self.setImg2)
        # self.player1.sigend.connect(self.set_player1)
        # self.player2.sigend.connect(self.set_player2)
        self.record_num = 0
        self.show_image()

    def set_autoSlider(self):
        if self.imageArea:
            self.timer_slider.stop()
            if self.slider_auto.value() > 0:
                self.imageArea.set_auoMode(self.btnGroup_auto.checkedId() + 1)
                self.timer_slider.start(1000 / (self.slider_auto.value()))
            else:
                self.imageArea.set_auoMode(0)

    def auto_slider(self):
        self.imageArea.auto_slider()

    def add_image(self, cb, btn):
        if self.btnGroup_mode.checkedId() == 0:
            path = QtWidgets.QFileDialog.getExistingDirectory(self)
            if len(path) <= 0:
                return
            cb.clear()
            files = sorted(os.listdir(path))
            for f in files:
                cb.addItem(path + '/' + f)
        else:
            if btn == 'add_1':
                path, _ = QtWidgets.QFileDialog.getOpenFileName(
                    self, 'Select gt file', '', 'Images (*.jpg *.png)')
                if len(path) <= 0:
                    return
                files = [path]
            elif btn == 'add_2':
                paths = QtWidgets.QFileDialog.getExistingDirectory(self)
                if len(paths) <= 0:
                    return
                files = sorted(os.listdir(paths))
                files = [paths + '/' + f for f in files]

            for path in files:
                if self.cb_1.count() > 0:
                    if self.cb_1.findText(path) > -1:
                        self.cb_1.removeItem(self.cb_1.findText(path))
                    if self.cb_2.findText(path) > -1:
                        self.cb_2.removeItem(self.cb_2.findText(path))
                    self.cb_1.addItem(path)
                    self.cb_2.addItem(path)
                    self.cb_2.setCurrentIndex(self.cb_2.count() - 1)
                else:
                    self.cb_1.addItem(path)
                    self.cb_2.addItem(path)
                    self.cb_1.setCurrentIndex(0)
                    self.cb_2.setCurrentIndex(0)

    def add_1(self):
        if self.btnGroup_type.checkedId() == 1:
            self.add_image(self.cb_1, 'add_1')
        else:
            path, _ = QtWidgets.QFileDialog.getOpenFileName(
                self, 'Select gt file', '', 'Images (*.mp4 *.avi)')
            if len(path) <= 0:
                return
            if self.cb_1.count() > 0:
                if self.cb_1.findText(path) > -1:
                    self.cb_1.removeItem(self.cb_1.findText(path))
                if self.cb_2.findText(path) > -1:
                    self.cb_2.removeItem(self.cb_2.findText(path))
                self.cb_1.addItem(path)
                self.cb_2.addItem(path)
                self.cb_2.setCurrentIndex(self.cb_2.count() - 1)
            else:
                self.cb_1.addItem(path)
                self.cb_2.addItem(path)
                self.cb_1.setCurrentIndex(0)
                self.cb_2.setCurrentIndex(0)

    def add_2(self):
        if self.btnGroup_type.checkedId() == 1:
            self.add_image(self.cb_2, 'add_2')
        else:
            path = QtWidgets.QFileDialog.getExistingDirectory(self)
            if len(path) <= 0:
                return
            if self.cb_1.count() > 0:
                if self.cb_1.findText(path) > -1:
                    self.cb_1.removeItem(self.cb_1.findText(path))
                if self.cb_2.findText(path) > -1:
                    self.cb_2.removeItem(self.cb_2.findText(path))
                self.cb_1.addItem(path)
                self.cb_2.addItem(path)
                self.cb_2.setCurrentIndex(self.cb_2.count() - 1)
            else:
                self.cb_1.addItem(path)
                self.cb_2.addItem(path)
                self.cb_1.setCurrentIndex(0)
                self.cb_2.setCurrentIndex(0)

    def set_label(self):
        if self.imageArea is not None:
            self.imageArea.label_1 = self.input_label_1.text()
            self.imageArea.label_2 = self.input_label_2.text()
            self.imageArea.title = self.input_title.text()
            self.imageArea.update()

    def set_scale(self):
        """Set scale."""
        self.scale = self.slider_scale.value()
        self.txt_scale.setText(f'{self.scale} %')
        if self.imageArea is not None:
            self.imageArea.set_scale(self.scale / 100.0)

    def change_image_1(self):
        if self.btnGroup_type.checkedId() == 0:
            self.change_video()
        else:
            if self.btnGroup_mode.checkedId() == 0:
                self.cb_2.setCurrentIndex(self.cb_1.currentIndex())
            self.images[0] = cv2.imread(self.cb_1.currentText())
            self.images[1] = cv2.imread(self.cb_2.currentText())
            self.show_image()

    def change_image_2(self):
        if self.btnGroup_type.checkedId() == 0:
            self.change_video()
        else:
            if self.btnGroup_mode.checkedId() == 0:
                self.cb_1.setCurrentIndex(self.cb_2.currentIndex())
            self.images[0] = cv2.imread(self.cb_1.currentText())
            self.images[1] = cv2.imread(self.cb_2.currentText())
            self.show_image()

    def change_video(self):
        if self.cb_1.currentText() != '' and self.cb_2.currentText() != '':
            self.player.set(self.cb_1.currentText(), self.cb_2.currentText())
            self.player.start()
            self.btn_pause.setEnabled(True)

    # def setImg1(self, img):
    #     self.images[0] = img
    #     self.imageArea.setImage(self.images)
    # def setImg2(self, img):
    #     self.images[1] = img
    #     self.imageArea.setImage(self.images)
    def setImg(self, images):
        self.images = images
        self.imageArea.setImage(self.images)

    def show_image(self):
        self.imageArea = QLabelSlider(self,
                                      self.slider_scale.value() / 100.0 + 1e-7,
                                      self.input_label_1.text(),
                                      self.input_label_2.text(),
                                      self.input_title.text())
        self.imageArea.setFrameShape(QtWidgets.QFrame.Box)
        self.imageArea.setLineWidth(2)
        self.imageArea.setAlignment(QtCore.Qt.AlignBottom)
        self.imageArea.setStyleSheet(
            'border-width: 0px; border-style: solid; border-color: rgb(100, 100, 100);background-color: rgb(255, 255, 255)'  # noqa
        )
        self.image_scroll.setWidget(self.imageArea)

    def pause(self):
        if self.btn_pause.text() == 'Pause (Space)':
            self.player.pause()
            self.timer_slider.stop()
            self.btn_pause.setText('Play (Space)')
        else:
            self.player.resume()
            self.set_autoSlider()
            self.btn_pause.setText('Pause (Space)')

    def reset(self):
        self.images = [None, None]
        self.cb_1.clear()
        self.cb_2.clear()
        self.show_image()
        if self.btnGroup_type.checkedId() == 1:
            if self.btnGroup_mode.checkedId() == 0:
                self.btn_add_1.setText('Set image 1')
                self.btn_add_2.setText('Set image 2')
            else:
                self.btn_add_1.setText('Add file')
                self.btn_add_2.setText('Add directory')
        else:
            self.btn_add_1.setText('Add a video')
            self.btn_add_2.setText('Add frames')

        self.btn_pause.setText('Pause (Space)')
        self.btn_pause.setEnabled(False)
        self.btn_record.setText('Record (Enter)')

        self.timer_slider.stop()
        self.timer_record.stop()

    def save(self):
        """Save slider compare result."""
        path, _ = QtWidgets.QFileDialog.getSaveFileName(self, 'save')
        if len(path) <= 0:
            return
        if self.imageArea:
            self.imageArea.grab().save(path)
            QtWidgets.QMessageBox.about(self, 'Message', 'Success!')
        else:
            QtWidgets.QMessageBox.about(self, 'Message', 'Nothing to save.')

    def record(self):
        if self.btn_record.text() == 'Record (Enter)':
            self.record_num = 0
            self.timer_record.start(1000 / 25)
            self.txt_prompt.setText('Recording...')
            self.btn_record.setText('End (Enter)')
        elif self.btn_record.text() == 'End (Enter)':
            paths = sorted(os.listdir('.tmp/'))
            paths = ['.tmp/' + p for p in paths]
            if len(paths) <= 0:
                return
            img = cv2.imread(paths[0])
            h, w, _ = img.shape
            cur = datetime.datetime.now()
            fname = f'{cur.year}{cur.month}{cur.day}{cur.hour}{cur.minute}{cur.second}.mp4'  # noqa
            fourcc = cv2.VideoWriter_fourcc(*'mp4v')
            self.recorder = cv2.VideoWriter(fname, fourcc, 25, (w, h))
            self.txt_prompt.setText('Saving...')
            for i in range(self.record_num):
                img = cv2.imread(paths[i])
                img = cv2.resize(img, (w, h))
                self.recorder.write(img)
                os.remove(paths[i])
            self.recorder.release()
            self.timer_record.stop()
            self.txt_prompt.setText('')
            self.btn_record.setText('Record (Enter)')
            QtWidgets.QMessageBox.about(self, 'Message',
                                        f'Save {fname} success!')

    def recording(self):
        if not os.path.isdir('.tmp/'):
            os.makedirs('.tmp/')
        fname = '.tmp/' + str(self.record_num).zfill(8) + '.png'
        self.record_num += 1
        self.imageArea.grab().save(fname)

    def wheelEvent(self, ev) -> None:
        key = QtWidgets.QApplication.keyboardModifiers()
        if key == QtCore.Qt.ControlModifier:
            scale = ev.angleDelta().y() / 120
            self.scale += scale
            if self.scale < 1:
                self.scale = 1
            self.txt_scale.setText(f'{self.scale} %')
            if self.scale > 200 and self.imageArea is not None:
                self.imageArea.set_scale(self.scale / 100.0)
            else:
                self.slider_scale.setValue(self.scale)
        return super().wheelEvent(ev)

    def keyPressEvent(self, ev) -> None:
        if ev.key() == QtCore.Qt.Key_Left:
            if self.cb_1.currentIndex() > 0:
                self.cb_1.setCurrentIndex(self.cb_1.currentIndex() - 1)
            else:
                self.cb_1.setCurrentIndex(self.cb_1.count() - 1)
        elif ev.key() == QtCore.Qt.Key_Right:
            if self.cb_1.currentIndex() < self.cb_1.count() - 1:
                self.cb_1.setCurrentIndex(self.cb_1.currentIndex() + 1)
            else:
                self.cb_1.setCurrentIndex(0)
        elif ev.key(
        ) == QtCore.Qt.Key_Enter or ev.key() + 1 == QtCore.Qt.Key_Enter:
            self.record()
        elif ev.key() == QtCore.Qt.Key_Space:
            self.pause()

    def eventFilter(self, object, event) -> bool:
        if object == self.image_scroll or object == self:
            if event.type() == QtCore.QEvent.KeyPress:
                self.keyPressEvent(event)
                return True
        return super().eventFilter(object, event)


class SRPage(QtWidgets.QWidget):

    def __init__(self, parent) -> None:
        super().__init__()

        self.parent = parent
        self.statusBar = self.parent.statusBar

        self.tab_patch = PatchTab(self)
        self.tab_slider = SliderTab(self)
        self.tabs = QtWidgets.QTabWidget()
        self.tabs.addTab(self.tab_patch, 'patch compare')
        self.tabs.addTab(self.tab_slider, 'before/after slider')

        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.tabs)
        # self.tabs.currentChanged.connect(self.tabsCurrentChanged)



# DistributedDataParallel
model_wrapper_cfg = dict(type='MMSeparateDistributedDataParallel')

model = dict(
    type='TwoStageInpaintor',
    disc_input_with_mask=True,
    data_preprocessor=dict(
        type='DataPreprocessor',
        mean=[127.5],
        std=[127.5],
    ),
    encdec=dict(
        type='DeepFillEncoderDecoder',
        stage1=dict(
            type='GLEncoderDecoder',
            encoder=dict(
                type='DeepFillEncoder',
                conv_type='gated_conv',
                channel_factor=0.75,
                padding_mode='reflect'),
            decoder=dict(
                type='DeepFillDecoder',
                conv_type='gated_conv',
                in_channels=96,
                channel_factor=0.75,
                out_act_cfg=dict(type='Tanh'),
                padding_mode='reflect'),
            dilation_neck=dict(
                type='GLDilationNeck',
                in_channels=96,
                conv_type='gated_conv',
                act_cfg=dict(type='ELU'),
                padding_mode='reflect')),
        stage2=dict(
            type='DeepFillRefiner',
            encoder_attention=dict(
                type='DeepFillEncoder',
                encoder_type='stage2_attention',
                conv_type='gated_conv',
                channel_factor=0.75,
                padding_mode='reflect'),
            encoder_conv=dict(
                type='DeepFillEncoder',
                encoder_type='stage2_conv',
                conv_type='gated_conv',
                channel_factor=0.75,
                padding_mode='reflect'),
            dilation_neck=dict(
                type='GLDilationNeck',
                in_channels=96,
                conv_type='gated_conv',
                act_cfg=dict(type='ELU'),
                padding_mode='reflect'),
            contextual_attention=dict(
                type='ContextualAttentionNeck',
                in_channels=96,
                conv_type='gated_conv',
                padding_mode='reflect'),
            decoder=dict(
                type='DeepFillDecoder',
                in_channels=192,
                conv_type='gated_conv',
                out_act_cfg=dict(type='Tanh'),
                padding_mode='reflect'))),
    disc=dict(
        type='MultiLayerDiscriminator',
        in_channels=4,
        max_channels=256,
        fc_in_channels=None,
        num_convs=6,
        norm_cfg=None,
        act_cfg=dict(type='LeakyReLU', negative_slope=0.2),
        out_act_cfg=dict(type='LeakyReLU', negative_slope=0.2),
        with_spectral_norm=True,
    ),
    stage1_loss_type=('loss_l1_hole', 'loss_l1_valid'),
    stage2_loss_type=('loss_l1_hole', 'loss_l1_valid', 'loss_gan'),
    loss_gan=dict(
        type='GANLoss',
        gan_type='hinge',
        loss_weight=0.1,
    ),
    loss_l1_hole=dict(
        type='L1Loss',
        loss_weight=1.0,
    ),
    loss_l1_valid=dict(
        type='L1Loss',
        loss_weight=1.0,
    ),
)

# optimizer
optim_wrapper = dict(
    constructor='MultiOptimWrapperConstructor',
    generator=dict(
        type='OptimWrapper', optimizer=dict(type='Adam', lr=0.0001)),
    disc=dict(type='OptimWrapper', optimizer=dict(type='Adam', lr=0.0001)))

# learning policy
# Fixed



# Copyright (c) OpenMMLab. All rights reserved.
from mmengine import MessageHub
from mmengine.optim import _ParamScheduler

from mmagic.registry import PARAM_SCHEDULERS


@PARAM_SCHEDULERS.register_module()
class ReduceLR(_ParamScheduler):
    """Decays the learning rate of each parameter group by linearly changing
    small multiplicative factor until the number of epoch reaches a pre-defined
    milestone: ``end``.

    Notice that such decay can happen simultaneously with other changes to the
    learning rate from outside this scheduler.

    Note:
        The learning rate of each parameter group will be update at regular
            intervals.

    Args:
        optimizer (Optimizer or OptimWrapper): Wrapped optimizer.
        mode (str, optional): One of `min`, `max`. In `min` mode, lr will
            be reduced when the quantity monitored has stopped
            decreasing; in `max` mode it will be reduced when the
            quantity monitored has stopped increasing. Default: 'min'.
        factor (float, optional): Factor by which the learning rate will be
            reduced. new_lr = lr * factor. Default: 0.1.
        patience (int, optional): Number of epochs with no improvement after
            which learning rate will be reduced. For example, if
            `patience = 2`, then we will ignore the first 2 epochs
            with no improvement, and will only decrease the LR after the
            3rd epoch if the loss still hasn't improved then.
            Default: 10.
        threshold (float, optional): Threshold for measuring the new optimum,
            to only focus on significant changes. Default: 1e-4.
        threshold_mode (str, optional): One of `rel`, `abs`. In `rel` mode,
            dynamic_threshold = best * ( 1 + threshold ) in 'max'
            mode or best * ( 1 - threshold ) in `min` mode.
            In `abs` mode, dynamic_threshold = best + threshold in
            `max` mode or best - threshold in `min` mode. Default: 'rel'.
        cooldown (int, optional): Number of epochs to wait before resuming
            normal operation after lr has been reduced. Default: 0.
        min_lr (float, optional): Minimum LR value to keep. If LR after decay
            is lower than `min_lr`, it will be clipped to this value.
            Default: 0.
        eps (float, optional): Minimal decay applied to lr. If the difference
            between new and old lr is smaller than eps, the update is
            ignored. Default: 1e-8.
        begin (int): Step at which to start updating the learning rate.
            Defaults to 0.
        end (int): Step at which to stop updating the learning rate.
        last_step (int): The index of last step. Used for resume without
            state dict. Defaults to -1.
        by_epoch (bool): Whether the scheduled learning rate is updated by
            epochs. Defaults to True.
    """

    def __init__(self,
                 optimizer,
                 mode: str = 'min',
                 factor: float = 0.1,
                 patience: int = 10,
                 threshold: float = 1e-4,
                 threshold_mode: str = 'rel',
                 cooldown: int = 0,
                 min_lr: float = 0.,
                 eps: float = 1e-8,
                 **kwargs):

        super().__init__(optimizer=optimizer, param_name='lr', **kwargs)

        self.message_hub = MessageHub.get_instance('reduce_lr')

        if mode not in ['min', 'max']:
            raise ValueError(
                'mode must be one of "min" or "max", instead got {mode}')
        self.mode = mode

        if factor >= 1.0 or factor < 0:
            raise ValueError('Factor should be < 1.0 and >=0')
        self.factor = factor

        self.patience = patience
        self.threshold = threshold

        if threshold_mode not in ['rel', 'abs']:
            raise ValueError('thresh_mode must be one of "rel" or "abs",'
                             f'instead got {threshold_mode}')
        self.threshold_mode = threshold_mode

        self.cooldown = cooldown
        self.cooldown_counter = 0
        self.best = None
        self.num_bad_epochs = None
        self.mode_worse = None  # the worse value for the chosen mode
        self.min_lr = min_lr
        self.eps = eps
        self.last_epoch = 0
        self._init_is_better(self.mode)
        self._reset()

    def _get_value(self):
        """Compute value using chainable form of the scheduler."""

        if self.last_step == 0:
            return [
                group[self.param_name] for group in self.optimizer.param_groups
            ]

        current = self.message_hub.get_scalar('value').current()
        if self.is_better(current, self.best):
            self.best = current
            self.num_bad_epochs = 0
        else:
            self.num_bad_epochs += 1

        if self.in_cooldown:
            self.cooldown_counter -= 1
            self.num_bad_epochs = 0

        if self.num_bad_epochs > self.patience:
            self.cooldown_counter = self.cooldown
            self.num_bad_epochs = 0
            results = []
            for group in self.optimizer.param_groups:
                regular_lr = group[self.param_name]
                if regular_lr - regular_lr * self.factor > self.eps:
                    regular_lr = max(regular_lr * self.factor, self.min_lr)
                results.append(regular_lr)
            return results

        else:
            return [
                group[self.param_name] for group in self.optimizer.param_groups
            ]

    def _init_is_better(self, mode):
        if mode == 'min':
            self.mode_worse = float('inf')
        else:
            self.mode_worse = float('-inf')

    def _reset(self):
        self.best = self.mode_worse
        self.cooldown_counter = 0
        self.num_bad_epochs = 0

    def is_better(self, a, best):
        if self.mode == 'min' and self.threshold_mode == 'rel':
            rel_epsilon = 1. - self.threshold
            return a < best * rel_epsilon
        elif self.mode == 'min' and self.threshold_mode == 'abs':
            return a < best - self.threshold
        elif self.mode == 'max' and self.threshold_mode == 'rel':
            rel_epsilon = 1. + self.threshold
            return a > best * rel_epsilon
        else:
            return a > best + self.threshold

    @property
    def in_cooldown(self):
        return self.cooldown_counter > 0



test_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='img',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(type='PackInputs')
]

# test config for Set5
set5_data_root = 'data/Set5'
set5_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='set5', task_name='sisr'),
        data_root=set5_data_root,
        data_prefix=dict(img='LRbicx4', gt='GTmod12'),
        pipeline=test_pipeline))
set5_evaluator = dict(
    type='Evaluator',
    metrics=[
        dict(type='PSNR', crop_border=4, prefix='Set5'),
        dict(type='SSIM', crop_border=4, prefix='Set5'),
    ])

set14_data_root = 'data/Set14'
set14_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        metainfo=dict(dataset_type='set14', task_name='sisr'),
        data_root=set14_data_root,
        data_prefix=dict(img='LRbicx4', gt='GTmod12'),
        pipeline=test_pipeline))
set14_evaluator = dict(
    type='Evaluator',
    metrics=[
        dict(type='PSNR', crop_border=4, prefix='Set14'),
        dict(type='SSIM', crop_border=4, prefix='Set14'),
    ])

# test config for DIV2K
div2k_data_root = 'data/DIV2K'
div2k_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicImageDataset',
        ann_file='meta_info_DIV2K100sub_GT.txt',
        metainfo=dict(dataset_type='div2k', task_name='sisr'),
        data_root=div2k_data_root,
        data_prefix=dict(
            img='DIV2K_train_LR_bicubic/X4_sub', gt='DIV2K_train_HR_sub'),
        pipeline=test_pipeline))
div2k_evaluator = dict(
    type='Evaluator',
    metrics=[
        dict(type='PSNR', crop_border=4, prefix='DIV2K'),
        dict(type='SSIM', crop_border=4, prefix='DIV2K'),
    ])

# test config
test_cfg = dict(type='MultiTestLoop')
test_dataloader = [
    set5_dataloader,
    set14_dataloader,
    div2k_dataloader,
]
test_evaluator = [
    set5_evaluator,
    set14_evaluator,
    div2k_evaluator,
]



# Copyright (c) OpenMMLab. All rights reserved.
from .hooks import *  # noqa: F401, F403
from .optimizers import *  # noqa: F401, F403
from .runner import *  # noqa: F401, F403
from .schedulers import *  # noqa: F401, F403



# Copyright (c) OpenMMLab. All rights reserved.
from mmengine.model import BaseModule

from mmagic.registry import MODELS


@MODELS.register_module()
class PConvEncoderDecoder(BaseModule):
    """Encoder-Decoder with partial conv module.

    Args:
        encoder (dict): Config of the encoder.
        decoder (dict): Config of the decoder.
    """

    def __init__(self, encoder, decoder):
        super().__init__()
        self.encoder = MODELS.build(encoder)
        self.decoder = MODELS.build(decoder)

        # support fp16
        self.fp16_enabled = False

    def forward(self, x, mask_in):
        """Forward Function.

        Args:
            x (torch.Tensor): Input tensor with shape of (n, c, h, w).
            mask_in (torch.Tensor): Input tensor with shape of (n, c, h, w).

        Returns:
            torch.Tensor: Output tensor with shape of (n, c, h', w').
        """
        enc_outputs = self.encoder(x, mask_in)
        x, final_mask = self.decoder(enc_outputs)

        return x, final_mask



_base_ = [
    '../_base_/default_runtime.py',
    '../_base_/datasets/basicvsr_test_config.py'
]

experiment_name = 'basicvsr_2xb4_reds4'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs'

scale = 4

# model settings
model = dict(
    type='BasicVSR',
    generator=dict(
        type='BasicVSRNet',
        mid_channels=64,
        num_blocks=30,
        spynet_pretrained='https://download.openmmlab.com/mmediting/restorers/'
        'basicvsr/spynet_20210409-c6c1bd09.pth'),
    pixel_loss=dict(type='CharbonnierLoss', loss_weight=1.0, reduction='mean'),
    train_cfg=dict(fix_iter=5000),
    data_preprocessor=dict(
        type='DataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
    ))

train_pipeline = [
    dict(type='GenerateSegmentIndices', interval_list=[1]),
    dict(type='LoadImageFromFile', key='img', channel_order='rgb'),
    dict(type='LoadImageFromFile', key='gt', channel_order='rgb'),
    dict(type='SetValues', dictionary=dict(scale=scale)),
    dict(type='PairedRandomCrop', gt_patch_size=256),
    dict(
        type='Flip',
        keys=['img', 'gt'],
        flip_ratio=0.5,
        direction='horizontal'),
    dict(
        type='Flip', keys=['img', 'gt'], flip_ratio=0.5, direction='vertical'),
    dict(type='RandomTransposeHW', keys=['img', 'gt'], transpose_ratio=0.5),
    dict(type='PackInputs')
]

val_pipeline = [
    dict(type='GenerateSegmentIndices', interval_list=[1]),
    dict(type='LoadImageFromFile', key='img', channel_order='rgb'),
    dict(type='LoadImageFromFile', key='gt', channel_order='rgb'),
    dict(type='PackInputs')
]

demo_pipeline = [
    dict(type='GenerateSegmentIndices', interval_list=[1]),
    dict(type='LoadImageFromFile', key='img', channel_order='rgb'),
    dict(type='PackInputs')
]

data_root = 'data/REDS'

train_dataloader = dict(
    num_workers=6,
    batch_size=4,
    persistent_workers=False,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    dataset=dict(
        type='BasicFramesDataset',
        metainfo=dict(dataset_type='reds_reds4', task_name='vsr'),
        data_root=data_root,
        data_prefix=dict(img='train_sharp_bicubic/X4', gt='train_sharp'),
        ann_file='meta_info_reds4_train.txt',
        depth=1,
        num_input_frames=15,
        pipeline=train_pipeline))

val_dataloader = dict(
    num_workers=1,
    batch_size=1,
    persistent_workers=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type='BasicFramesDataset',
        metainfo=dict(dataset_type='reds_reds4', task_name='vsr'),
        data_root=data_root,
        data_prefix=dict(img='train_sharp_bicubic/X4', gt='train_sharp'),
        ann_file='meta_info_reds4_val.txt',
        depth=1,
        num_input_frames=100,
        fixed_seq_len=100,
        pipeline=val_pipeline))

val_evaluator = dict(
    type='Evaluator', metrics=[
        dict(type='PSNR'),
        dict(type='SSIM'),
    ])

train_cfg = dict(
    type='IterBasedTrainLoop', max_iters=300_000, val_interval=5000)
val_cfg = dict(type='MultiValLoop')

# optimizer
optim_wrapper = dict(
    constructor='DefaultOptimWrapperConstructor',
    type='OptimWrapper',
    optimizer=dict(type='Adam', lr=2e-4, betas=(0.9, 0.99)),
    paramwise_cfg=dict(custom_keys={'spynet': dict(lr_mult=0.125)}))

default_hooks = dict(checkpoint=dict(out_dir=save_dir))

# learning policy
param_scheduler = dict(
    type='CosineRestartLR',
    by_epoch=False,
    periods=[300000],
    restart_weights=[1],
    eta_min=1e-7)

find_unused_parameters = True



_base_ = [
    '../_base_/default_runtime.py', '../_base_/datasets/sisr_x4_test_config.py'
]

experiment_name = 'edsr_x4c64b16_1xb16-300k_div2k'
work_dir = f'./work_dirs/{experiment_name}'
save_dir = './work_dirs/'

load_from = None  # based on pre-trained x2 model

scale = 4
# model settings
model = dict(
    type='BaseEditModel',
    generator=dict(
        type='EDSRNet',
        in_channels=3,
        out_channels=3,
        mid_channels=64,
        num_blocks=16,
        upscale_factor=scale,
        res_scale=1,
        rgb_mean=[0.4488, 0.4371, 0.4040],
        rgb_std=[1.0, 1.0, 1.0]),
    pixel_loss=dict(type='L1Loss', loss_weight=1.0, reduction='mean'),
    train_cfg=dict(),
    test_cfg=dict(metrics=['PSNR'], crop_border=scale),
    data_preprocessor=dict(
        type='DataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
    ))

train_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='img',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(type='SetValues', dictionary=dict(scale=scale)),
    dict(type='PairedRandomCrop', gt_patch_size=196),
    dict(
        type='Flip',
        keys=['img', 'gt'],
        flip_ratio=0.5,
        direction='horizontal'),
    dict(
        type='Flip', keys=['img', 'gt'], flip_ratio=0.5, direction='vertical'),
    dict(type='RandomTransposeHW', keys=['img', 'gt'], transpose_ratio=0.5),
    dict(type='PackInputs')
]
val_pipeline = [
    dict(
        type='LoadImageFromFile',
        key='img',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(
        type='LoadImageFromFile',
        key='gt',
        color_type='color',
        channel_order='rgb',
        imdecode_backend='cv2'),
    dict(type='PackInputs')
]

# dataset settings
dataset_type = 'BasicImageDataset'
data_root = 'data'

train_dataloader = dict(
    num_workers=8,
    batch_size=16,
    drop_last=True,
    persistent_workers=False,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    dataset=dict(
        type=dataset_type,
        ann_file='meta_info_DIV2K800sub_GT.txt',
        metainfo=dict(dataset_type='div2k', task_name='sisr'),
        data_root=data_root + '/DIV2K',
        data_prefix=dict(
            img='DIV2K_train_LR_bicubic/X4_sub', gt='DIV2K_train_HR_sub'),
        filename_tmpl=dict(img='{}', gt='{}'),
        pipeline=train_pipeline))

val_dataloader = dict(
    num_workers=4,
    persistent_workers=False,
    drop_last=False,
    sampler=dict(type='DefaultSampler', shuffle=False),
    dataset=dict(
        type=dataset_type,
        metainfo=dict(dataset_type='set5', task_name='sisr'),
        data_root=data_root + '/Set5',
        data_prefix=dict(img='LRbicx4', gt='GTmod12'),
        pipeline=val_pipeline))

val_evaluator = dict(
    type='Evaluator',
    metrics=[
        dict(type='MAE'),
        dict(type='PSNR', crop_border=scale),
        dict(type='SSIM', crop_border=scale),
    ])

train_cfg = dict(
    type='IterBasedTrainLoop', max_iters=300000, val_interval=5000)
val_cfg = dict(type='MultiValLoop')

# optimizer
optim_wrapper = dict(
    constructor='DefaultOptimWrapperConstructor',
    type='OptimWrapper',
    optimizer=dict(type='Adam', lr=1e-4, betas=(0.9, 0.999)))

# learning policy
param_scheduler = dict(
    type='MultiStepLR', by_epoch=False, milestones=[200000], gamma=0.5)

default_hooks = dict(
    checkpoint=dict(
        type='CheckpointHook',
        interval=5000,
        save_optimizer=True,
        by_epoch=False,
        out_dir=save_dir,
    ),
    timer=dict(type='IterTimerHook'),
    logger=dict(type='LoggerHook', interval=100),
    param_scheduler=dict(type='ParamSchedulerHook'),
    sampler_seed=dict(type='DistSamplerSeedHook'),
)



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os.path as osp
import re

import cv2
import mmcv
import mmengine
import numpy as np

from mmagic.evaluation import gauss_gradient
from mmagic.utils import modify_args


def sad(alpha, trimap, pred_alpha):
    if alpha.ndim != 2 or trimap.ndim != 2 or pred_alpha.ndim != 2:
        raise ValueError(
            'input alpha, trimap and pred_alpha should has two dimensions, '
            f'alpha {alpha.shape}, please check their shape: '
            f'trimap {trimap.shape}, pred_alpha {pred_alpha.shape}')
    assert (pred_alpha[trimap == 0] == 0).all()
    assert (pred_alpha[trimap == 255] == 255).all()
    alpha = alpha.astype(np.float64) / 255
    pred_alpha = pred_alpha.astype(np.float64) / 255
    sad_result = np.abs(pred_alpha - alpha).sum() / 1000
    return sad_result


def mse(alpha, trimap, pred_alpha):
    if alpha.ndim != 2 or trimap.ndim != 2 or pred_alpha.ndim != 2:
        raise ValueError(
            'input alpha, trimap and pred_alpha should has two dimensions, '
            f'alpha {alpha.shape}, please check their shape: '
            f'trimap {trimap.shape}, pred_alpha {pred_alpha.shape}')
    assert (pred_alpha[trimap == 0] == 0).all()
    assert (pred_alpha[trimap == 255] == 255).all()
    alpha = alpha.astype(np.float64) / 255
    pred_alpha = pred_alpha.astype(np.float64) / 255
    weight_sum = (trimap == 128).sum()
    if weight_sum != 0:
        mse_result = ((pred_alpha - alpha)**2).sum() / weight_sum
    else:
        mse_result = 0
    return mse_result


def gradient_error(alpha, trimap, pred_alpha, sigma=1.4):
    """Gradient error for evaluating alpha matte prediction.

    Args:
        alpha (ndarray): Ground-truth alpha matte.
        trimap (ndarray): Input trimap with its value in {0, 128, 255}.
        pred_alpha (ndarray): Predicted alpha matte.
        sigma (float): Standard deviation of the gaussian kernel. Default: 1.4.
    """
    if alpha.ndim != 2 or trimap.ndim != 2 or pred_alpha.ndim != 2:
        raise ValueError(
            'input alpha, trimap and pred_alpha should has two dimensions, '
            f'alpha {alpha.shape}, please check their shape: '
            f'trimap {trimap.shape}, pred_alpha {pred_alpha.shape}')
    if not ((pred_alpha[trimap == 0] == 0).all() and
            (pred_alpha[trimap == 255] == 255).all()):
        raise ValueError(
            'pred_alpha should be masked by trimap before evaluation')
    alpha = alpha.astype(np.float64)
    pred_alpha = pred_alpha.astype(np.float64)
    alpha_normed = np.zeros_like(alpha)
    pred_alpha_normed = np.zeros_like(pred_alpha)
    cv2.normalize(alpha, alpha_normed, 1., 0., cv2.NORM_MINMAX)
    cv2.normalize(pred_alpha, pred_alpha_normed, 1., 0., cv2.NORM_MINMAX)

    alpha_grad = gauss_gradient(alpha_normed, sigma).astype(np.float32)
    pred_alpha_grad = gauss_gradient(pred_alpha_normed,
                                     sigma).astype(np.float32)

    grad_loss = ((alpha_grad - pred_alpha_grad)**2 * (trimap == 128)).sum()
    # same as SAD, divide by 1000 to reduce the magnitude of the result
    return grad_loss / 1000


def connectivity(alpha, trimap, pred_alpha, step=0.1):
    """Connectivity error for evaluating alpha matte prediction.

    Args:
        alpha (ndarray): Ground-truth alpha matte with shape (height, width).
            Value range of alpha is [0, 255].
        trimap (ndarray): Input trimap with shape (height, width). Elements
            in trimap are one of {0, 128, 255}.
        pred_alpha (ndarray): Predicted alpha matte with shape (height, width).
            Value range of pred_alpha is [0, 255].
        step (float): Step of threshold when computing intersection between
            `alpha` and `pred_alpha`.
    """
    if alpha.ndim != 2 or trimap.ndim != 2 or pred_alpha.ndim != 2:
        raise ValueError(
            'input alpha, trimap and pred_alpha should has two dimensions, '
            f'alpha {alpha.shape}, please check their shape: '
            f'trimap {trimap.shape}, pred_alpha {pred_alpha.shape}')
    if not ((pred_alpha[trimap == 0] == 0).all() and
            (pred_alpha[trimap == 255] == 255).all()):
        raise ValueError(
            'pred_alpha should be masked by trimap before evaluation')
    alpha = alpha.astype(np.float32) / 255
    pred_alpha = pred_alpha.astype(np.float32) / 255

    thresh_steps = np.arange(0, 1 + step, step)
    round_down_map = -np.ones_like(alpha)
    for i in range(1, len(thresh_steps)):
        alpha_thresh = alpha >= thresh_steps[i]
        pred_alpha_thresh = pred_alpha >= thresh_steps[i]
        intersection = (alpha_thresh & pred_alpha_thresh).astype(np.uint8)

        # connected components
        _, output, stats, _ = cv2.connectedComponentsWithStats(
            intersection, connectivity=4)
        # start from 1 in dim 0 to exclude background
        size = stats[1:, -1]

        # largest connected component of the intersection
        omega = np.zeros_like(alpha)
        if len(size) != 0:
            max_id = np.argmax(size)
            # plus one to include background
            omega[output == max_id + 1] = 1

        mask = (round_down_map == -1) & (omega == 0)
        round_down_map[mask] = thresh_steps[i - 1]
    round_down_map[round_down_map == -1] = 1

    alpha_diff = alpha - round_down_map
    pred_alpha_diff = pred_alpha - round_down_map
    # only calculate difference larger than or equal to 0.15
    alpha_phi = 1 - alpha_diff * (alpha_diff >= 0.15)
    pred_alpha_phi = 1 - pred_alpha_diff * (pred_alpha_diff >= 0.15)

    connectivity_error = np.sum(
        np.abs(alpha_phi - pred_alpha_phi) * (trimap == 128))
    # same as SAD, divide by 1000 to reduce the magnitude of the result
    return connectivity_error / 1000


def evaluate_one(args):
    """Function to evaluate one sample of data.

    Args:
        args (tuple): Information needed to evaluate one sample of data.

    Returns:
        dict: The evaluation results including sad, mse, gradient error and
            connectivity error.
    """
    pred_alpha_path, alpha_path, trimap_path = args
    pred_alpha = mmcv.imread(pred_alpha_path, flag='grayscale')
    alpha = mmcv.imread(alpha_path, flag='grayscale')
    if trimap_path is None:
        trimap = np.ones_like(alpha)
    else:
        trimap = mmcv.imread(trimap_path, flag='grayscale')
    sad_result = sad(alpha, trimap, pred_alpha)
    mse_result = mse(alpha, trimap, pred_alpha)
    grad_result = gradient_error(alpha, trimap, pred_alpha)
    conn_result = connectivity(alpha, trimap, pred_alpha)
    return (sad_result, mse_result, grad_result, conn_result)


def evaluate(pred_root, gt_root, trimap_root, verbose, nproc):
    """Evaluate test results of Adobe composition-1k dataset.

    There are 50 different ground truth foregrounds and alpha mattes pairs,
    each of the foreground will be composited with 20 different backgrounds,
    producing 1000 images for testing. In some repo, the ground truth alpha
    matte will be copied 20 times and named the same as the images. This
    function accept both original alpha matte folder (contains 50 ground
    truth alpha mattes) and copied alpha matte folder (contains 1000 ground
    truth alpha mattes) for `gt_root`.

    Example of copied name:
    ```
    alpha_matte1.png -> alpha_matte1_0.png
                        alpha_matte1_1.png
                        ...
                        alpha_matte1_19.png
                        alpha_matte1_20.png
    ```

    Args:
        pred_root (str): Path to the predicted alpha matte folder.
        gt_root (str): Path to the ground truth alpha matte folder.
        trimap_root (str): Path to the predicted alpha matte folder.
        verbose (bool): Whether print result for each predicted alpha matte.
        nproc (int): number of processes.
    """

    images = sorted(mmengine.scandir(pred_root))
    gt_files_num = len(list(mmengine.scandir(gt_root)))
    # If ground truth alpha mattes are not copied (number of files is 50), we
    # use the below pattern to recover the name of the original alpha matte.
    if gt_files_num == 50:
        pattern = re.compile(r'(.+)_(?:\d+)(.png)')
    pairs = []
    for img in images:
        pred_alpha_path = osp.join(pred_root, img)
        # if ground truth alpha matte are not copied, recover the original name
        if gt_files_num == 50:
            groups = pattern.match(img).groups()
            alpha_path = osp.join(gt_root, ''.join(groups))
        # if ground truth alpha matte are copied, the name should be the same
        else:  # gt_files_num == 1000
            alpha_path = osp.join(gt_root, img)
        trimap_path = (
            osp.join(trimap_root, img) if trimap_root is not None else None)
        pairs.append((pred_alpha_path, alpha_path, trimap_path))

    results = mmengine.track_parallel_progress(evaluate_one, pairs, nproc)

    if verbose:
        # for sad_result, mse_result, grad_result, conn_result in results:
        for i, img in enumerate(images):
            sad_result, mse_result, grad_result, conn_result = results[i]
            print(f'{img} SAD: {sad_result:.6g} MattingMSE: {mse_result:.6g} '
                  f'GRAD: {grad_result:.6g} CONN: {conn_result:.6g}')

    sad_mean, mse_mean, grad_mean, conn_mean = np.mean(results, axis=0)
    print(f'MEAN:  SAD: {sad_mean:.6g} MattingMSE: {mse_mean:.6g} '
          f'GRAD: {grad_mean:.6g} CONN: {conn_mean:.6g}')


def parse_args():
    modify_args()
    parser = argparse.ArgumentParser(
        description='evaluate composition-1k prediction result')
    parser.add_argument(
        'pred_root', help='Path to the predicted alpha matte folder')
    parser.add_argument(
        'gt_root', help='Path to the ground truth alpha matte folder')
    parser.add_argument(
        '--trimap-root',
        help='Path to trimap folder. If not specified, '
        'results are calculated on the full image.')
    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        help='Whether print result for each predicted alpha matte')
    parser.add_argument(
        '--nproc', type=int, default=4, help='number of processes')
    return parser.parse_args()


def main():
    args = parse_args()

    if not osp.exists(args.pred_root):
        raise FileNotFoundError(f'pred_root {args.pred_root} not found')
    if not osp.exists(args.gt_root):
        raise FileNotFoundError(f'gt_root {args.gt_root} not found')

    evaluate(args.pred_root, args.gt_root, args.trimap_root, args.verbose,
             args.nproc)


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
# Adapted from https://github.com/huggingface/diffusers/blob/main/
# src/diffusers/models/resnet.py

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange


class InflatedConv3d(nn.Conv2d):
    """An implementation of InflatedConv3d."""

    def forward(self, x):
        """forward function."""
        video_length = x.shape[2]

        x = rearrange(x, 'b c f h w -> (b f) c h w')
        x = super().forward(x)
        x = rearrange(x, '(b f) c h w -> b c f h w', f=video_length)

        return x


class InflatedGroupNorm(nn.GroupNorm):

    def forward(self, x):
        video_length = x.shape[2]

        x = rearrange(x, 'b c f h w -> (b f) c h w')
        x = super().forward(x)
        x = rearrange(x, '(b f) c h w -> b c f h w', f=video_length)

        return x


class Upsample3D(nn.Module):
    """An 3D upsampling layer with an optional convolution.

    Args:
        channels (int): channels in the inputs and outputs.
        use_conv (bool): a bool determining if a convolution is applied.
        use_conv_transpose (bool): whether to use conv transpose.
        out_channels (int): output channels.
    """

    def __init__(self,
                 channels,
                 use_conv=False,
                 use_conv_transpose=False,
                 out_channels=None,
                 name='conv'):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.use_conv_transpose = use_conv_transpose
        self.name = name

        if use_conv_transpose:
            raise NotImplementedError
        elif use_conv:
            self.conv = InflatedConv3d(
                self.channels, self.out_channels, 3, padding=1)

    def forward(self, hidden_states, output_size=None):
        """forward with hidden states."""
        assert hidden_states.shape[1] == self.channels

        if self.use_conv_transpose:
            raise NotImplementedError

        # Cast to float32 to as 'upsample_nearest2d_out_frame'
        # op does not support bfloat16
        dtype = hidden_states.dtype
        if dtype == torch.bfloat16:
            hidden_states = hidden_states.to(torch.float32)

        # upsample_nearest_nhwc fails with large batch sizes.
        # see https://github.com/huggingface/diffusers/issues/984
        if hidden_states.shape[0] >= 64:
            hidden_states = hidden_states.contiguous()

        # if `output_size` is passed we force the interpolation output
        # size and do not make use of `scale_factor=2`
        if output_size is None:
            hidden_states = F.interpolate(
                hidden_states, scale_factor=[1.0, 2.0, 2.0], mode='nearest')
        else:
            hidden_states = F.interpolate(
                hidden_states, size=output_size, mode='nearest')

        # If the input is bfloat16, we cast back to bfloat16
        if dtype == torch.bfloat16:
            hidden_states = hidden_states.to(dtype)

        # if self.use_conv:
        #     if self.name == "conv":
        #         hidden_states = self.conv(hidden_states)
        #     else:
        #         hidden_states = self.Conv2d_0(hidden_states)
        hidden_states = self.conv(hidden_states)

        return hidden_states


class Downsample3D(nn.Module):
    """A 3D downsampling layer with an optional convolution.

    Args:
        channels (int): channels in the inputs and outputs.
        use_conv (bool): a bool determining if a convolution is applied.
        out_channels (int): output channels
        padding (int): padding num
    """

    def __init__(self,
                 channels,
                 use_conv=False,
                 out_channels=None,
                 padding=1,
                 name='conv'):
        super().__init__()
        self.channels = channels
        self.out_channels = out_channels or channels
        self.use_conv = use_conv
        self.padding = padding
        stride = 2
        self.name = name

        if use_conv:
            self.conv = InflatedConv3d(
                self.channels,
                self.out_channels,
                3,
                stride=stride,
                padding=padding)
        else:
            raise NotImplementedError

    def forward(self, hidden_states):
        """forward with hidden states."""
        assert hidden_states.shape[1] == self.channels
        if self.use_conv and self.padding == 0:
            raise NotImplementedError

        assert hidden_states.shape[1] == self.channels
        hidden_states = self.conv(hidden_states)

        return hidden_states


class ResnetBlock3D(nn.Module):
    """3D resnet block support down sample and up sample.

    Args:
        in_channels (int): input channels.
        out_channels (int): output channels.
        conv_shortcut (bool): whether to use conv shortcut.
        dropout (float): dropout rate.
        temb_channels (int): time embedding channels.
        groups (int): conv groups.
        groups_out (int): conv out groups.
        pre_norm (bool): whether to norm before conv. Todo: remove.
        eps (float): eps for groupnorm.
        non_linearity (str): non linearity type.
        time_embedding_norm (str): time embedding norm type.
        output_scale_factor (float): factor to scale input and output.
        use_in_shortcut (bool): whether to use conv in shortcut.
    """

    def __init__(
        self,
        *,
        in_channels,
        out_channels=None,
        conv_shortcut=False,
        dropout=0.0,
        temb_channels=512,
        groups=32,
        groups_out=None,
        pre_norm=True,
        eps=1e-6,
        non_linearity='swish',
        time_embedding_norm='default',
        output_scale_factor=1.0,
        use_in_shortcut=None,
        use_inflated_groupnorm=None,
    ):
        super().__init__()
        self.pre_norm = pre_norm
        self.pre_norm = True
        self.in_channels = in_channels
        out_channels = in_channels if out_channels is None else out_channels
        self.out_channels = out_channels
        self.use_conv_shortcut = conv_shortcut
        self.time_embedding_norm = time_embedding_norm
        self.output_scale_factor = output_scale_factor

        if groups_out is None:
            groups_out = groups

        if use_inflated_groupnorm:
            self.norm1 = InflatedGroupNorm(
                num_groups=groups,
                num_channels=in_channels,
                eps=eps,
                affine=True)
        else:
            self.norm1 = torch.nn.GroupNorm(
                num_groups=groups,
                num_channels=in_channels,
                eps=eps,
                affine=True)

        self.conv1 = InflatedConv3d(
            in_channels, out_channels, kernel_size=3, stride=1, padding=1)

        if temb_channels is not None:
            if self.time_embedding_norm == 'default':
                time_emb_proj_out_channels = out_channels
            elif self.time_embedding_norm == 'scale_shift':
                time_emb_proj_out_channels = out_channels * 2
            else:
                raise ValueError(f'unknown time_embedding_norm : '
                                 f'{self.time_embedding_norm} ')

            self.time_emb_proj = torch.nn.Linear(temb_channels,
                                                 time_emb_proj_out_channels)
        else:
            self.time_emb_proj = None

        if use_inflated_groupnorm:
            self.norm2 = InflatedGroupNorm(
                num_groups=groups_out,
                num_channels=out_channels,
                eps=eps,
                affine=True)
        else:
            self.norm2 = torch.nn.GroupNorm(
                num_groups=groups_out,
                num_channels=out_channels,
                eps=eps,
                affine=True)

        self.dropout = torch.nn.Dropout(dropout)
        self.conv2 = InflatedConv3d(
            out_channels, out_channels, kernel_size=3, stride=1, padding=1)

        if non_linearity == 'swish':
            self.nonlinearity = lambda x: F.silu(x)
        elif non_linearity == 'mish':
            self.nonlinearity = Mish()
        elif non_linearity == 'silu':
            self.nonlinearity = nn.SiLU()

        self.use_in_shortcut = self.in_channels != self.out_channels \
            if use_in_shortcut is None else use_in_shortcut

        self.conv_shortcut = None
        if self.use_in_shortcut:
            self.conv_shortcut = InflatedConv3d(
                in_channels, out_channels, kernel_size=1, stride=1, padding=0)

    def forward(self, input_tensor, temb):
        """forward with hidden states and time embeddings."""
        hidden_states = input_tensor

        hidden_states = self.norm1(hidden_states)
        hidden_states = self.nonlinearity(hidden_states)

        hidden_states = self.conv1(hidden_states)

        if temb is not None:
            temb = self.time_emb_proj(self.nonlinearity(temb))[:, :, None,
                                                               None, None]

        if temb is not None and self.time_embedding_norm == 'default':
            hidden_states = hidden_states + temb

        hidden_states = self.norm2(hidden_states)

        if temb is not None and self.time_embedding_norm == 'scale_shift':
            scale, shift = torch.chunk(temb, 2, dim=1)
            hidden_states = hidden_states * (1 + scale) + shift

        hidden_states = self.nonlinearity(hidden_states)

        hidden_states = self.dropout(hidden_states)
        hidden_states = self.conv2(hidden_states)

        if self.conv_shortcut is not None:
            input_tensor = self.conv_shortcut(input_tensor)

        output_tensor = (input_tensor +
                         hidden_states) / self.output_scale_factor

        return output_tensor


class Mish(torch.nn.Module):
    """Mish activation function."""

    def forward(self, hidden_states):
        """forward with hidden states."""
        return hidden_states * torch.tanh(
            torch.nn.functional.softplus(hidden_states))
